#!python
#cython: language_level=3
"""
Implements the sequence edit distance of Levenshtein (1965) and its
backtracing in cython.

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import random
import heapq
import numpy as np
from cython.parallel import prange
from libc.math cimport sqrt
cimport cython
from edist.alignment import Alignment

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'


###################################
# Edit Distance with Custom Delta #
###################################

def sed(x, y, delta = None):
    """ Computes the sequence edit distance between the input sequence
    x and the input sequence y, given the element-wise distance function delta.

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.
    delta: function (default = None)
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them. If None, this
        method calls standard_sed instead.

    Returns
    -------
    d: float
        the sequence edit distance between x and y according to delta.

    """
    if(delta is None):
        return float(standard_sed(x, y))
    _, _, _, D = _sed(x, y, delta)
    return D[0,0]

def _sed(x, y, delta):
    """ Internal function. Call sed instead. """
    cdef int m = len(x)
    cdef int n = len(y)
    # First, compute all pairwise replacements
    Delta = np.zeros((m, n))
    cdef double[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            Delta_view[i,j] = delta(x[i], y[j])

    # Then, compute all deletions
    Delta_del = np.zeros(m)
    cdef double[:] Delta_del_view = Delta_del
    for i in range(m):
        Delta_del_view[i] = delta(x[i], None)

    # Then, compute all insertions
    Delta_ins = np.zeros(n)
    cdef double[:] Delta_ins_view = Delta_ins
    for j in range(n):
        Delta_ins_view[j] = delta(None, y[j])

    # Then, compute the sequence edit distance
    D = np.zeros((m+1,n+1))
    sed_c(Delta, Delta_del, Delta_ins, D)

    return Delta, Delta_del, Delta_ins, D


@cython.boundscheck(False)
cdef void sed_c(const double[:,:] Delta, const double[:] Delta_del, const double[:] Delta_ins, double[:,:] D) nogil:
    """ Computes the sequence edit distance between two input sequences
    with pairwise element distances Delta and an (empty) dynamic programming
    matrix D.

    Arguments
    ---------
    Delta: double matrix
        a m x n matrix containing the pairwise element distances.
    Delta_del: double array
        a m-element vector containing deletion costs.
    Delta_ins: double array
        a n-element vector containing insertion costs.
    D: double matrix
        an m+1 x n+1 matrix to which the output will be written.
        The sequence edit distance will be in cell [0, 0] after the computation
        is finished.

    """
    cdef int m = Delta.shape[0]
    cdef int n = Delta.shape[1]
    cdef int i
    cdef int j
    # initialize last entry
    D[m, n] = 0.
    # compute last column
    for i in range(m-1,-1,-1):
        D[i,n] = Delta_del[i] + D[i+1,n]
    # compute last row
    for j in range(n-1,-1,-1):
        D[m,j] = Delta_ins[j] + D[m,j+1]
    # compute remaining matrix
    for i in range(m-1,-1,-1):
        for j in range(n-1,-1,-1):
            D[i,j] = min3(Delta[i,j] + D[i+1,j+1],
                          Delta_del[i] + D[i+1, j],
                          Delta_ins[j] + D[i, j+1])

cdef double min3(double a, double b, double c) nogil:
    """ Computes the minimum of three numbers.

    Parameters
    ----------
    a: double
        a number
    b: double
        another number
    c: double
        yet another number

    Returns
    -------
    min3: double
        min({a, b, c})

    """
    if(a < b):
        if(a < c):
            return a
        else:
            return c
    else:
        if(b < c):
            return b
        else:
            return c

#########################
# Backtracing Functions #
#########################

cdef double _BACKTRACE_TOL = 1E-5

def sed_backtrace(x, y, delta = None):
    """ Computes a co-optimal alignment between the two input sequences
    x and y, given the element-wise distance function delta. This mechanism
    is deterministic and will always prefer replacements over other options.

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.
    delta: function (default = None)
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them. If None, this
        method calls standard_sed_backtrace instead.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the sequence edit
        distance.

    """
    if(delta is None):
        return standard_sed_backtrace(x, y)
    cdef int m = len(x)
    cdef int n = len(y)
    Delta, Delta_del, Delta_ins, D = _sed(x, y, delta)
    cdef double[:,:] Delta_view = Delta
    cdef double[:] Delta_del_view = Delta_del
    cdef double[:] Delta_ins_view = Delta_ins
    cdef double[:,:] D_view = D

    # Finally, compute the backtrace
    cdef int i = 0
    cdef int j = 0
    alignment = Alignment()
    while(i < m and j < n):
        # check which alignment option is co-optimal
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j+1]):
            # replacement is co-optimal
            alignment.append_tuple(i, j)
            i += 1
            j += 1
            continue
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_del_view[i] + D_view[i+1,j]):
            # deleting x[i] is co-optimal
            alignment.append_tuple(i, -1)
            i += 1
            continue
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_ins_view[j] + D_view[i,j+1]):
            # inserting y[j] is co-optimal
            alignment.append_tuple(-1, j)
            j += 1
            continue
        # if we got here, nothing is co-optimal, which is an error
        raise ValueError('Internal error: No option is co-optimal.')
    while(i < m):
        alignment.append_tuple(i, -1)
        i += 1
    while(j < n):
        alignment.append_tuple(-1, j)
        j += 1
    return alignment

def sed_backtrace_stochastic(x, y, delta = None):
    """ Computes a co-optimal alignment between the two input sequences
    x and y, given the element-wise distance function delta. This mechanism
    is stochastic and will return a random co-optimal alignment.

    Note that the randomness does _not_ produce a uniform distribution over
    all co-optimal alignments because reandom choices at the start of the
    alignment process dominate. If you wish to characterize the overall
    distribution accurately, use sed_backtrace_matrix instead.

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.
    delta: function (default = None)
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them. If None, this
        method calls standard_sed_backtrace_stochastic instead.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the sequence edit
        distance.

    """
    if(delta is None):
        return standard_sed_backtrace_stochastic(x, y)
    cdef int m = len(x)
    cdef int n = len(y)
    Delta, Delta_del, Delta_ins, D = _sed(x, y, delta)
    cdef double[:,:] Delta_view = Delta
    cdef double[:] Delta_del_view = Delta_del
    cdef double[:] Delta_ins_view = Delta_ins
    cdef double[:,:] D_view = D

    # Finally, compute the backtrace
    cdef int i = 0
    cdef int j = 0
    alignment = Alignment()
    while(i < m and j < n):
        # check which alignment option is co-optimal
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j+1]):
            if(D_view[i,j] + _BACKTRACE_TOL > Delta_del_view[i] + D_view[i+1,j]):
                if(D_view[i,j] + _BACKTRACE_TOL > Delta_ins_view[j] + D_view[i,j+1]):
                    # replacement, deletion, and insertion is co-optimal
                    # Select how to proceed uniformly at random
                    r = random.randrange(3)
                    if(r == 0):
                        alignment.append_tuple(i, j)
                        i += 1
                        j += 1
                    elif(r == 1):
                        alignment.append_tuple(i, -1)
                        i += 1
                    else:
                        alignment.append_tuple(-1, j)
                        j += 1
                else:
                    # replacement and deletion is co-optimal.
                    # Select how to proceed uniformly at random
                    if(random.randrange(2) == 0):
                        alignment.append_tuple(i, j)
                        i += 1
                        j += 1
                    else:
                        alignment.append_tuple(i, -1)
                        i += 1
            elif(D_view[i,j] + _BACKTRACE_TOL > Delta_ins_view[j] + D_view[i,j+1]):
                # replacement and insertion is co-optimal.
                # Select how to proceed uniformly at random
                if(random.randrange(2) == 0):
                    alignment.append_tuple(i, j)
                    i += 1
                    j += 1
                else:
                    alignment.append_tuple(-1, j)
                    j += 1
            else:
                # only replacement is co-optimal
                alignment.append_tuple(i, j)
                i += 1
                j += 1
        elif(D_view[i,j] + _BACKTRACE_TOL > Delta_del_view[i] + D_view[i+1,j]):
            if(D_view[i,j] + _BACKTRACE_TOL > Delta_ins_view[j] + D_view[i,j+1]):
                # deletion and insertion is cooptimal
                # Select how to proceed uniformly at random
                if(random.randrange(2) == 0):
                    alignment.append_tuple(i, -1)
                    i += 1
                else:
                    alignment.append_tuple(-1, j)
                    j += 1
            else:
                # only deletion is co-optimal
                alignment.append_tuple(i, -1)
                i += 1
        elif(D_view[i,j] + _BACKTRACE_TOL > Delta_ins_view[j] + D_view[i,j+1]):
            # only insertion is co-optimal
            alignment.append_tuple(-1, j)
            j += 1
        else:
            # if we got here, nothing is co-optimal, which is an error
            raise ValueError('Internal error: No option is co-optimal.')
    while(i < m):
        alignment.append_tuple(i, -1)
        i += 1
    while(j < n):
        alignment.append_tuple(-1, j)
        j += 1
    return alignment

def sed_backtrace_matrix(x, y, delta = None):
    """ Computes a matrix, summarizing all co-optimal alignments between
    x and y in a matrix P, where entry P[i, j] specifies the fraction of
    co-optimal alignments in which node x[i] has been aligned with node y[j].

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.
    delta: function (default = None)
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them. If None, this
        method calls standard_sed_backtrace_matrix instead.

    Returns
    -------
    P: array_like
        a matrix, where entry P[i, j] specifies the fraction of co-optimal
        alignments in which node x[i] has been aligned with node y[j].
        P[i, n] contains the fraction of deletions of node x[i] and P[m, j]
        the fraction of insertions of node y[j].
    K: array_like
        a matrix that contains the counts for all co-optimal alignments in
        which node x[i] has been aligned with node y[j].
    k: int
        the number of co-optimal alignments overall, such that P = K / k.

    """
    if(delta is None):
        return standard_sed_backtrace_matrix(x, y)
    cdef int m = len(x)
    cdef int n = len(y)
    Delta, Delta_del, Delta_ins, D = _sed(x, y, delta)
    cdef double[:,:] Delta_view = Delta
    cdef double[:] Delta_del_view = Delta_del
    cdef double[:] Delta_ins_view = Delta_ins
    cdef double[:,:] D_view = D

    # compute the forward matrix Alpha, which contains the number of
    # co-optimal alignment paths from cell [0, 0] to cell [i, j]
    Alpha = np.zeros((m+1, n+1), dtype=int)
    cdef long[:,:] Alpha_view = Alpha
    Alpha_view[0, 0] = 1
    # build a queue of cells which we still need to process
    q = [(0, 0)]
    # build a set which stores the already visited cells
    visited = set()
    # initialize temporary variables
    cdef int found_coopt = False
    cdef int i
    cdef int j
    cdef long k = 0
    while(q):
        (i, j) = heapq.heappop(q)
        if((i, j) in visited):
            continue
        visited.add((i, j))
        k = Alpha_view[i, j]
        if(i == m):
            if(j == n):
                continue
            # if we are at the end of the first sequence, we can only insert
            Alpha_view[i, j+1] += k
            heapq.heappush(q, (i, j+1))
            continue
        if(j == n):
            # if we are at the end of the second sequence, we can only delete
            Alpha_view[i+1, j] += k
            heapq.heappush(q, (i+1, j))
            continue
        found_coopt = False
        # check which alignment option is co-optimal
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j+1]):
            # replacement is co-optimal
            Alpha_view[i+1, j+1] += k
            heapq.heappush(q, (i+1, j+1))
            found_coopt = True
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_del_view[i] + D_view[i+1,j]):
            # deletion is co-optimal
            Alpha_view[i+1, j] += k
            heapq.heappush(q, (i+1, j))
            found_coopt = True
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_ins_view[j] + D_view[i,j+1]):
            # insertion is co-optimal
            Alpha_view[i, j+1] += k
            heapq.heappush(q, (i, j+1))
            found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    # compute the backward matrix Beta, which contains the number of
    # co-optimal alignment paths from cell [i, j] to cell [m, n]
    Beta = np.zeros((m+1, n+1), dtype=int)
    cdef long[:,:] Beta_view = Beta
    Beta_view[m, n] = 1
    # iterate in downward lexigraphic order over the visited cells
    for (i, j) in sorted(visited, reverse = True):
        k = Beta_view[i, j]
        if(i == 0):
            if(j == 0):
                continue
            # if we are at the start of the first sequence, we can only insert
            Beta_view[i, j-1] += k
            continue
        if(j == 0):
            # if we are at the start of the second sequence, we can only delete
            Beta_view[i-1, j] += k
            continue
        found_coopt = False
        # check which alignment option is co-optimal
        if(D_view[i-1,j-1] + _BACKTRACE_TOL > Delta_view[i-1,j-1] + D_view[i,j]):
            # replacement is co-optimal
            Beta_view[i-1, j-1] += k
            found_coopt = True
        if(D_view[i-1,j] + _BACKTRACE_TOL > Delta_del_view[i-1] + D_view[i,j]):
            # deletion is co-optimal
            Beta_view[i-1, j] += k
            found_coopt = True
        if(D_view[i,j-1] + _BACKTRACE_TOL > Delta_ins_view[j-1] + D_view[i,j]):
            # insertion is co-optimal
            Beta_view[i, j-1] += k
            found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    if(Alpha_view[m, n] != Beta_view[0, 0]):
        raise ValueError('Internal error: Alignment count in Alpha and Beta matrix did not agree; got %d versus %d' % (Alpha_view[m, n], Beta_view[0, 0]))

    # compute a counting matrix specifying how often each alignment has
    # occured by multiplying alpha and beta values.
    K = np.zeros((m, n), dtype=int)
    cdef long[:,:] K_view = K
    for (i, j) in visited:
        if(i == m or j == n):
            continue
        # check if replacement is co-optimal
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j+1]):
            K_view[i, j] = Alpha_view[i, j] * Beta_view[i+1, j+1]

    # compute the final summary matrix by dividing K by the overall number
    # of co-optimal alignments and completing the last row and column
    k = Alpha_view[m, n]
    P = np.zeros((m+1, n+1))
    P[:m, :][:, :n] = K
    P[:m, n] = k - np.sum(K, axis=1)
    P[m, :n] = k - np.sum(K, axis=0)
    P /= k

    return P, K, k

###############################################
# Standard Edit Distance with Kronecker Delta #
###############################################

def standard_sed(x, y):
    """ Computes the standard sequence edit distance/Levenshtein distance
    between the input sequence x and the input sequence y.

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.

    Returns
    -------
    d: int
        the standard sequence edit distance between x and y.

    """
    _, D = _standard_sed(x, y)
    return D[0, 0]

def _standard_sed(x, y):
    """ Internal function. Call standard_sed instead. """
    cdef int m = len(x)
    cdef int n = len(y)
    # First, compute all pairwise replacements
    Delta = np.zeros((m, n), dtype=int)
    cdef long[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            if(x[i] != y[j]):
                Delta_view[i, j] = 1

    # Then, compute the sequence edit distance
    D = np.zeros((m+1,n+1), dtype=int)
    standard_sed_c(Delta, D)
    return Delta, D

@cython.boundscheck(False)
def sed_string(str x, str y):
    """ Computes the standard sequence edit distance/Levenshtein distance
    between two input strings x and y, using the Kronecker distance as
    element-wise distance measure.

    Parameters
    ----------
    x: str
        a string.
    y: str
        another string.

    Returns
    -------
    d: int
        the standard sequence edit distance between x and y.

    """
    cdef int m = len(x)
    cdef int n = len(y)
    # First, compute all pairwise replacements
    Delta = np.zeros((m, n), dtype=int)
    cdef long[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in prange(m, nogil=True):
        for j in prange(n):
            if(x[i] != y[j]):
                Delta_view[i, j] = 1

    # Then, compute the standard sequence edit distance
    D = np.zeros((m+1,n+1), dtype=int)
    standard_sed_c(Delta, D)
    return D[0,0]

@cython.boundscheck(False)
cdef void standard_sed_c(const long[:,:] Delta, long[:,:] D) nogil:
    """ Computes the standard sequence edit distance between two input sequences
    with pairwise element distances Delta and an (empty) dynamic programming
    matrix D.

    Parameters
    ----------
    Delta: long matrix
        a m x n matrix containing the pairwise element distances.
    D: long matrix
        another m x n matrix to which the output will be written.
        The sequence edit distance will be in cell [0, 0] after the computation
        is finished.

    """
    cdef int m = Delta.shape[0]
    cdef int n = Delta.shape[1]
    cdef int i
    cdef int j
    # initialize last entry
    D[m, n] = 0
    # compute last column
    for i in range(m-1,-1,-1):
        D[i,n] = 1 + D[i+1,n]
    # compute last row
    for j in range(n-1,-1,-1):
        D[m,j] = 1 + D[m,j+1]
    # compute remaining matrix
    for i in range(m-1,-1,-1):
        for j in range(n-1,-1,-1):
            D[i,j] = min3_int(Delta[i,j] + D[i+1,j+1],
                          1 + D[i+1, j],
                          1 + D[i, j+1])

cdef long min3_int(long a, long b, long c) nogil:
    """ Computes the minimum of three numbers.

    Parameters
    ----------
    a: int
        a number
    b: int
        another number
    c: int
        yet another number

    Returns
    -------
    min3: int
        min({a, b, c})

    """
    if(a < b):
        if(a < c):
            return a
        else:
            return c
    else:
        if(b < c):
            return b
        else:
            return c

#########################
# Backtracing Functions #
#########################

def standard_sed_backtrace(x, y):
    """ Computes a co-optimal alignment between the two input sequences
    x and y. This mechanism is deterministic and will always prefer
    replacements over other options.

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the sequence edit
        distance.

    """
    cdef int m = len(x)
    cdef int n = len(y)
    # Compute the standard edit distance first
    Delta, D = _standard_sed(x, y)
    cdef long[:,:] Delta_view = Delta
    cdef long[:,:] D_view = D

    # Then, compute the backtrace
    cdef int i = 0
    cdef int j = 0
    alignment = Alignment()
    while(i < m and j < n):
        # check which alignment option is co-optimal
        if(D_view[i,j] == Delta_view[i,j] + D_view[i+1,j+1]):
            # replacement is co-optimal
            alignment.append_tuple(i, j)
            i += 1
            j += 1
            continue
        if(D_view[i,j] == 1 + D_view[i+1,j]):
            # deleting x[i] is co-optimal
            alignment.append_tuple(i, -1)
            i += 1
            continue
        if(D_view[i,j] == 1 + D_view[i,j+1]):
            # inserting y[j] is co-optimal
            alignment.append_tuple(-1, j)
            j += 1
            continue
        # if we got here, nothing is co-optimal, which is an error
        raise ValueError('Internal error: No option is co-optimal.')
    while(i < m):
        alignment.append_tuple(i, -1)
        i += 1
    while(j < n):
        alignment.append_tuple(-1, j)
        j += 1
    return alignment

def standard_sed_backtrace_stochastic(x, y):
    """ Computes a co-optimal alignment between the two input sequences
    x and y, given the element-wise distance function delta. This mechanism
    is stochastic and will return a random co-optimal alignment.

    Note that the randomness does _not_ produce a uniform distribution over
    all co-optimal alignments because reandom choices at the start of the
    alignment process dominate. If you wish to characterize the overall
    distribution accurately, use sed_backtrace_matrix instead.

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the sequence edit
        distance.

    """

    cdef int m = len(x)
    cdef int n = len(y)
    # Compute the standard edit distance first
    Delta, D = _standard_sed(x, y)
    cdef long[:,:] Delta_view = Delta
    cdef long[:,:] D_view = D

    # Then, compute the backtrace
    cdef int i = 0
    cdef int j = 0
    alignment = Alignment()
    while(i < m and j < n):
        # check which alignment option is co-optimal
        if(D_view[i,j] == Delta_view[i,j] + D_view[i+1,j+1]):
            if(D_view[i,j] == 1 + D_view[i+1,j]):
                if(D_view[i,j] == 1 + D_view[i,j+1]):
                    # replacement, deletion, and insertion is co-optimal
                    # Select how to proceed uniformly at random
                    r = random.randrange(3)
                    if(r == 0):
                        alignment.append_tuple(i, j)
                        i += 1
                        j += 1
                    elif(r == 1):
                        alignment.append_tuple(i, -1)
                        i += 1
                    else:
                        alignment.append_tuple(-1, j)
                        j += 1
                else:
                    # replacement and deletion is co-optimal.
                    # Select how to proceed uniformly at random
                    if(random.randrange(2) == 0):
                        alignment.append_tuple(i, j)
                        i += 1
                        j += 1
                    else:
                        alignment.append_tuple(i, -1)
                        i += 1
            elif(D_view[i,j] == 1 + D_view[i,j+1]):
                # replacement and insertion is co-optimal.
                # Select how to proceed uniformly at random
                if(random.randrange(2) == 0):
                    alignment.append_tuple(i, j)
                    i += 1
                    j += 1
                else:
                    alignment.append_tuple(-1, j)
                    j += 1
            else:
                # only replacement is co-optimal
                alignment.append_tuple(i, j)
                i += 1
                j += 1
        elif(D_view[i,j] == 1 + D_view[i+1,j]):
            if(D_view[i,j] == 1 + D_view[i,j+1]):
                # deletion and insertion is cooptimal
                # Select how to proceed uniformly at random
                if(random.randrange(2) == 0):
                    alignment.append_tuple(i, -1)
                    i += 1
                else:
                    alignment.append_tuple(-1, j)
                    j += 1
            else:
                # only deletion is co-optimal
                alignment.append_tuple(i, -1)
                i += 1
        elif(D_view[i,j] == 1 + D_view[i,j+1]):
            # only insertion is co-optimal
            alignment.append_tuple(-1, j)
            j += 1
        else:
            # if we got here, nothing is co-optimal, which is an error
            raise ValueError('Internal error: No option is co-optimal.')
    while(i < m):
        alignment.append_tuple(i, -1)
        i += 1
    while(j < n):
        alignment.append_tuple(-1, j)
        j += 1
    return alignment

def standard_sed_backtrace_matrix(x, y):
    """ Computes a matrix, summarizing all co-optimal alignments between
    x and y in a matrix P, where entry P[i, j] specifies the fraction of
    co-optimal alignments in which node x[i] has been aligned with node y[j].

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.

    Returns
    -------
    P: array_like
        a matrix, where entry P[i, j] specifies the fraction of co-optimal
        alignments in which node x[i] has been aligned with node y[j].
        P[i, n] contains the fraction of deletions of node x[i] and P[m, j]
        the fraction of insertions of node y[j].
    K: array_like
        a matrix that contains the counts for all co-optimal alignments in
        which node x[i] has been aligned with node y[j].
    k: int
        the number of co-optimal alignments overall, such that P = K / k.

    """

    cdef int m = len(x)
    cdef int n = len(y)
    # Compute the standard edit distance first
    Delta, D = _standard_sed(x, y)
    cdef long[:,:] Delta_view = Delta
    cdef long[:,:] D_view = D

    # compute the forward matrix Alpha, which contains the number of
    # co-optimal alignment paths from cell [0, 0] to cell [i, j]
    Alpha = np.zeros((m+1, n+1), dtype=int)
    cdef long[:,:] Alpha_view = Alpha
    Alpha_view[0, 0] = 1
    # build a queue of cells which we still need to process
    q = [(0, 0)]
    # build a set which stores the already visited cells
    visited = set()
    # initialize temporary variables
    cdef int found_coopt = False
    cdef int i
    cdef int j
    cdef long k = 0
    while(q):
        (i, j) = heapq.heappop(q)
        if((i, j) in visited):
            continue
        visited.add((i, j))
        k = Alpha_view[i, j]
        if(i == m):
            if(j == n):
                continue
            # if we are at the end of the first sequence, we can only insert
            Alpha_view[i, j+1] += k
            heapq.heappush(q, (i, j+1))
            continue
        if(j == n):
            # if we are at the end of the second sequence, we can only delete
            Alpha_view[i+1, j] += k
            heapq.heappush(q, (i+1, j))
            continue
        found_coopt = False
        # check which alignment option is co-optimal
        if(D_view[i,j] == Delta_view[i,j] + D_view[i+1,j+1]):
            # replacement is co-optimal
            Alpha_view[i+1, j+1] += k
            heapq.heappush(q, (i+1, j+1))
            found_coopt = True
        if(D_view[i,j] == 1 + D_view[i+1,j]):
            # deletion is co-optimal
            Alpha_view[i+1, j] += k
            heapq.heappush(q, (i+1, j))
            found_coopt = True
        if(D_view[i,j] == 1 + D_view[i,j+1]):
            # insertion is co-optimal
            Alpha_view[i, j+1] += k
            heapq.heappush(q, (i, j+1))
            found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    # compute the backward matrix Beta, which contains the number of
    # co-optimal alignment paths from cell [i, j] to cell [m, n]
    Beta = np.zeros((m+1, n+1), dtype=int)
    cdef long[:,:] Beta_view = Beta
    Beta_view[m, n] = 1
    # iterate in downward lexigraphic order over the visited cells
    for (i, j) in sorted(visited, reverse = True):
        k = Beta_view[i, j]
        if(i == 0):
            if(j == 0):
                continue
            # if we are at the start of the first sequence, we can only insert
            Beta_view[i, j-1] += k
            continue
        if(j == 0):
            # if we are at the start of the second sequence, we can only delete
            Beta_view[i-1, j] += k
            continue
        found_coopt = False
        # check which alignment option is co-optimal
        if(D_view[i-1,j-1] == Delta_view[i-1,j-1] + D_view[i,j]):
            # replacement is co-optimal
            Beta_view[i-1, j-1] += k
            found_coopt = True
        if(D_view[i-1,j] == 1 + D_view[i,j]):
            # deletion is co-optimal
            Beta_view[i-1, j] += k
            found_coopt = True
        if(D_view[i,j-1] == 1 + D_view[i,j]):
            # insertion is co-optimal
            Beta_view[i, j-1] += k
            found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    if(Alpha_view[m, n] != Beta_view[0, 0]):
        raise ValueError('Internal error: Alignment count in Alpha and Beta matrix did not agree; got %d versus %d' % (Alpha_view[m, n], Beta_view[0, 0]))

    # compute a counting matrix specifying how often each alignment has
    # occured by multiplying alpha and beta values.
    K = np.zeros((m, n), dtype=int)
    cdef long[:,:] K_view = K
    for (i, j) in visited:
        if(i == m or j == n):
            continue
        # check if replacement is co-optimal
        if(D_view[i,j] == Delta_view[i,j] + D_view[i+1,j+1]):
            K_view[i, j] = Alpha_view[i, j] * Beta_view[i+1, j+1]

    # compute the final summary matrix by dividing K by the overall number
    # of co-optimal alignments and completing the last row and column
    k = Alpha_view[m, n]
    P = np.zeros((m+1, n+1))
    P[:m, :][:, :n] = K
    P[:m, n] = k - np.sum(K, axis=1)
    P[m, :n] = k - np.sum(K, axis=0)
    P /= k

    return P, K, k




"""
Implements tree edits, i.e. functions which take a tree as input and
return a changed tree.

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import copy
from edist.ted import outermost_right_leaves

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

class Edit(abc.ABC):
    """ An abstract parent class for all edits.

    """
    @abc.abstractmethod
    def apply(self, nodes, adj):
        """ Applies this edit to the given tree and returns a copy of the tree
        with the applied changes. The original tree remains unchanged.

        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        Returns
        -------
        nodes: list
            a copy of nodes with the applied edit.
        adj: list
            a copy of adj with the applied edit.

        """
        pass

    @abc.abstractmethod
    def apply_in_place(self, nodes, adj):
        """ Applies this edit to the given tree. Note that this changes the
        input arguments.

        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        """
        pass


class Replacement(Edit):
    """ Replaces the label of node self._index with self._label.

    Attributes
    ----------
    _index: int
        The index of the tree node to which this edit is applied.
    _label: str
        The new label for the self._indexth node.

    """
    def __init__(self, index, label):
        self._index = index
        self._label = label

    def apply(self, nodes_orig, adj):
        """ Replaces the label of node self._index with self._label.

        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        Returns
        -------
        nodes: list
            a copy of nodes with the applied edit.
        adj: list
            a copy of adj with the applied edit.

        """
        nodes = copy.copy(nodes_orig)
        self.apply_in_place(nodes, adj)
        return nodes, adj

    def apply_in_place(self, nodes, adj):
        """ Replaces the label of node self._index with self._label.


        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        """
        nodes[self._index] = self._label

    def __repr__(self):
        return 'rep(%d, %s)' % (self._index, str(self._label))

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return isinstance(other, Replacement) and self._index == other._index and self._label == other._label


class Deletion(Edit):
    """ Deletes node self._index and raises all its children to the parent
        node. Note that deleting the root node of the tree results in a
        forest instead of a tree.

    Attributes
    ----------
    _index: int
        The index of the tree node to which this edit is applied.

    """
    def __init__(self, index):
        self._index = index

    def apply(self, nodes_orig, adj_orig):
        """ Deletes node self._index and raises all its children to the parent
        node. Note that deleting the root node of the tree results in a
        forest instead of a tree.

        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        Returns
        -------
        nodes: list
            a copy of nodes with the applied edit.
        adj: list
            a copy of adj with the applied edit.

        """
        nodes = copy.copy(nodes_orig)
        adj   = copy.deepcopy(adj_orig)
        self.apply_in_place(nodes, adj)
        return nodes, adj

    def apply_in_place(self, nodes, adj):
        """ Deletes node self._index and raises all its children to the parent
        node. Note that deleting the root node of the tree results in a
        forest instead of a tree.

        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        """
        # delete the nodes entry at index
        del nodes[self._index]
        # correct all adjacency list entries. First, we must raise
        # the children of index to the parent of index
        for i in range(len(adj)):
            if(self._index in adj[i]):
                j = adj[i].index(self._index)
                adj[i] = adj[i][:j] + adj[self._index] + adj[i][j+1:]
                break
        # next we must remove the index'th entry of the adjacency list
        del adj[self._index]
        # finally, we must decrement all entries in the adjacency list
        # bigger than index
        for i in range(len(adj)):
            for j in range(len(adj[i])):
                if(adj[i][j] > self._index):
                    adj[i][j] -= 1

    def __repr__(self):
        return 'del(%d)' % (self._index)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return isinstance(other, Deletion) and self._index == other._index


def get_roots(adj):
    """ Returns all roots of a forest, described by an adjacency list.

    Parameters
    ----------
    adj: list
        an adjacency list

    Returns
    -------
    roots: list
        a list of roots, ascendingly sorted.

    """
    # build a parent representation of the adjacency list
    p = [-1] * len(adj)
    for i in range(len(adj)):
        for j in adj[i]:
            p[j] = i
    # test which nodes have no parents; those are the roots
    roots = []
    for i in range(len(adj)):
        if(p[i] < 0):
            roots.append(i)
    return roots


class Insertion(Edit):
    """ Inserts a new node with label self._label as self._child_index'th
        child of node self._parent_index and uses the next self._num_children
        right siblings of itself as its children.

    Attributes
    ----------
    _parent_index: int
        The index of the tree node to which we add a new child.
    _label: str
        The label for the new child node.
    _child_index: int
        The new node will be the _child_indexth child of node _parent_index.
    _num_children: int
        The number of right siblings that will be used as new grandchildren.

    """
    def __init__(self, parent_index, child_index, label, num_children = 0):
        self._parent_index = parent_index
        self._label = label
        self._child_index = child_index
        self._num_children = num_children

    def apply(self, nodes_orig, adj_orig):
        """ Inserts a new node with label self._label as self._child_index'th
        child of node self._parent_index and uses the next self._num_children
        right siblings of itself as its children.

        Note that inserting a new child at the root leads to a forest instead
        of a tree structure.

        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        Returns
        -------
        nodes: list
            a copy of nodes with the applied edit.
        adj: list
            a copy of adj with the applied edit.

        """
        nodes = copy.copy(nodes_orig)
        adj   = copy.deepcopy(adj_orig)
        self.apply_in_place(nodes, adj)
        return nodes, adj

    def apply_in_place(self, nodes, adj):
        """ Inserts a new node with label self._label as self._child_index'th
        child of node self._parent_index and uses the next self._num_children
        right siblings of itself as its children.

        Note that inserting a new child at the root leads to a forest instead
        of a tree structure.

        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        """
        # introduce some abbreviations for easier typing
        p = self._parent_index
        c = self._child_index
        C = self._num_children

        if(p < 0):
            # first, handle the special case that we want to insert a new node
            # at the root of the tree, resulting in a forest
            # retrieve the current roots
            roots = get_roots(adj)
            # get the index of the current self._child_index'th root, which
            # will be the index of our new node
            if(c < len(roots)):
                idx = roots[c]
                # use the self._num_children next roots as children of our new
                # node
                children = roots[c:(c+C)]
            else:
                # if this child does not exist yet, append at the end
                idx = len(nodes)
                children = []
        else:
            # if our parent exists, get the index of its self._child_index'th
            # child, which will be the index of our new node
            if(c < len(adj[p])):
                idx = adj[p][c]
                # use the self._num_children right siblings as children of our new
                # node
                children = adj[p][c:(c+C)]
                del adj[p][c:(c+C)]
            else:
                # if this child does not exist yet, use the index of the
                # right sibling of the parent node
                # this index can be retrieved by first getting the outermost
                # right leaf of p.
                orl = p
                while(adj[orl]):
                    orl = adj[orl][-1]
                # and then incrementing it by 1
                idx = orl + 1
                children = []
            # insert the new child to the parent
            adj[p].insert(c, idx)
        # insert a new entry in the node list
        nodes.insert(idx, self._label)
        # insert a new entry into the adjacency list
        adj.insert(idx, children)
        # increment all indices >= idx
        for i in range(len(adj)):
            for j in range(len(adj[i])):
                if(adj[i][j] >= idx):
                    adj[i][j] += 1
        # decrement the actual new index again
        if(p >= 0):
            adj[p][c] -= 1

    def __repr__(self):
        return 'ins(%d, %d, %s, %d)' % (self._parent_index, self._child_index, self._label, self._num_children)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return isinstance(other, Insertion) and self._parent_index == other._parent_index and self._child_index == other._child_index and self._label == other._label and self._num_children == other._num_children


class Script(list, Edit):
    """ A list of Edits.

    """
    def __init__(self, lst = []):
        list.__init__(self, lst)

    def apply(self, nodes_orig, adj_orig):
        """ Applies all edits in this script.


        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        Returns
        -------
        nodes: list
            a copy of nodes with the applied edits.
        adj: list
            a copy of adj with the applied edits.

        """
        # if this script is empty, just return the inputs
        if(not self):
            return nodes_orig, adj_orig
        # otherwise make a copy and then apply the script in place
        nodes = copy.copy(nodes_orig)
        adj   = copy.deepcopy(adj_orig)
        self.apply_in_place(nodes, adj)
        return nodes, adj

    def apply_in_place(self, nodes, adj):
        """ Applies all edits in this script.

        Parameters
        ----------
        nodes: list
            a node list.
        adj: list
            an adjacency list.

        """
        for e in range(len(self)):
            self[e].apply_in_place(nodes, adj)


def alignment_to_script(alignment, x_nodes, x_adj, y_nodes, y_adj):
    """ Converts the given alignment into an edit script which maps the given
    tree x to the given tree y such that all tuples in the alignment
    correspond to exactly one edit in the script.

    Note that the order of operations does change because the script will
    first apply replacements (in input order), then deletions
    (in descending order), and finally insertions (in ascending order),
    which simplifies conversion.

    The precise algorithm is described in the paper:
    https://arxiv.org/abs/1805.06869

    Parameters
    ----------
    alignment: class alignment.Alignment
        an Alignment object which maps between the given trees x and y.
    x_nodes: list
        the node list of tree x.
    x_adj: list
        the adjacency list of tree x.
    y_nodes: list
        the node list of tree y.
    y_adj: list
        the adjacency list of tree y.

    Returns
    -------
    script: class tree_edits.Script
        A Script object script such that script.apply(x_nodes, x_adj) =
        (y_nodes, y_adj) where every Tuple in the alignment has one
        corresponding edit.

    """
    # in a first pass over the sequence, we disentangle replacements,
    # deletions, and insertions and convert the replacement right away because
    # they do not change the tree structure
    script = Script()
    dels = []
    inss = []
    for op in alignment:
        if(op._left >= 0):
            if(op._right >= 0):
                # if both left and right index are defined, we have a
                # a replacement that we can convert right away.
                # We ignore replacements that change nothing, though.
                if(x_nodes[op._left] != y_nodes[op._right]):
                    script.append(Replacement(op._left, y_nodes[op._right]))
            else:
                # if only the left index is defined, we have a deletion
                # which we will process later
                dels.append(op._left)
        else:
            # if only the right index is defined, we have an insertion
            # which we will process later
            inss.append(op._right)
    # in a next pass, we convert all deletions in descending order
    dels.reverse()
    dels.sort(reverse=True)
    for i in dels:
        script.append(Deletion(i))
    # the insertion conversion is, unfortunately, somewhat more complex.
    # For preparation, we need to sort the insertion indices ascendingly ...
    inss.sort()
    # then find, for each node, its parent index and its child index, ...
    ps = [-1] * len(y_nodes)
    cs = [-1] * len(y_nodes)
    cs[0] = 0
    for i in range(len(y_nodes)):
        for c in range(len(y_adj[i])):
            j = y_adj[i][c]
            cs[j] = c
            ps[j] = i
    # and, finally, the number of non-inserted descendants.
    Cs = num_descendants(y_adj, inss)
    # after this preparation, convert the insertions
    for i in inss:
        script.append(Insertion(ps[i], cs[i], y_nodes[i], Cs[i]))
    # return the script
    return script


def num_descendants(adj, filter_set):
    """ Counts the number of descendants of each node which are _not_ members
    of the given filter set.

    Parameters
    ----------
    adj: list
        an adjacency list
    filter_set: set_like
        a set excluding some node indices

    Returns
    -------
    out: list
        A list which contains, for each node, the number of descendants
        not from the filter set.

    """
    out = []
    _num_descendants(adj, filter_set, 0, out)
    return out

def _num_descendants(adj, filter_set, i, out):
    # add a counting variable for the number of descendants for the current node
    out.append(0)
    # iterate over all children of the current node
    for j in adj[i]:
        # call this method recursively for the child
        _num_descendants(adj, filter_set, j, out)
        # check if the current child is in the filter set
        if(j in filter_set):
            # if it is, add the number of descendants of the child
            out[i] += out[j]
        else:
            # otherwise, add only 1
            out[i] += 1



"""
Provides general utility functions to compute pairwise edit distances in
parallel.

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import multiprocessing as mp
import numpy as np

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

# The size of batches in parallel processing; this size has revealed
# to be favourable in empiric tests, even for longer sequences.
_BATCH_SIZE = 10

def _batch_dist_with_indices(k, l, dist, X, Y, symmetric = False):
    D = np.zeros((len(X), len(Y)))
    if(not symmetric):
        for k2 in range(len(X)):
            for l2 in range(len(Y)):
                D[k2, l2] = dist(X[k2], Y[l2])
    else:
        for k2 in range(len(X)):
            for l2 in range(k2+1, len(Y)):
                D[k2, l2] = dist(X[k2], Y[l2])
    return (k, l, D)

def _batch_dist_with_indices_and_delta(k, l, dist, X, Y, delta, symmetric = False):
    D = np.zeros((len(X), len(Y)))
    if(not symmetric):
        for k2 in range(len(X)):
            for l2 in range(len(Y)):
                D[k2, l2] = dist(X[k2], Y[l2], delta)
    else:
        for k2 in range(len(X)):
            for l2 in range(k2+1, len(Y)):
                D[k2, l2] = dist(X[k2], Y[l2], delta)
    return (k, l, D)

def pairwise_distances(Xs, Ys, dist, delta = None, num_jobs = 8):
    """ Computes the pairwise edit distances between the objects in
    Xs and the objects in Ys. Each object in Xs and Ys needs to be a valid
    input for the given distance function, i.e. a sequence or a tree.

    Optionally, it is possible to specify a component-wise distance function
    delta, which will then be forwarded to the input distance function

    Parameters
    ----------
    Xs: list
        a list of sequences or trees.
    Ys: list
        another list of sequences or trees.
    dist: function
        a function that takes an element of Xs as first and an element of
        Ys as second input and returns a scalar distance value between them.
    delta: function (default = None)
        a function that takes two elements of the input sequences or trees
        as inputs and returns their pairwise distance, where delta(x, None)
        should be the cost of deleting x and delta(None, y) should be the cost
        of inserting y. If this is not None, dist needs to accept an optional
        argument 'delta' as well. Defaults to None.
    num_jobs: int (default = 8)
        The number of jobs to be used for parallel processing. Defaults to 8.

    Returns
    -------
    D: array_like
        a len(Xs) x len(Ys) matrix of pairwise edit distance values.

    """
    K = len(Xs)
    L = len(Ys)
    # set up a parallel processing pool
    pool = mp.Pool(num_jobs)
    # set up the result matrix
    D = np.zeros((K,L))

    # start off all parallel processing jobs
    results = []
    if(delta is None):
        for k in range(0, K, _BATCH_SIZE):
            k_hi = min(k + _BATCH_SIZE, K)
            for l in range(0, L, _BATCH_SIZE):
                l_hi = min(l + _BATCH_SIZE, L)
                results.append(pool.apply_async(_batch_dist_with_indices, args=(k, l, dist, Xs[k:k_hi], Ys[l:l_hi])))
    else:
        for k in range(0, K, _BATCH_SIZE):
            k_hi = min(k + _BATCH_SIZE, K)
            for l in range(0, L, _BATCH_SIZE):
                l_hi = min(l + _BATCH_SIZE, L)
                results.append(pool.apply_async(_batch_dist_with_indices_and_delta, args=(k, l, dist, Xs[k:k_hi], Ys[l:l_hi], delta)))

    # close the pool
    pool.close()

    # sort all results into the matrix as soon as they arrive
    for res in results:
        k, l, D_batch = res.get()
        k_hi = k + D_batch.shape[0]
        l_hi = l + D_batch.shape[1]
        D[k:k_hi, :][:, l:l_hi] = D_batch

    # ensure that the entire pool is finished
    pool.join()

    # return the distance matrix
    return D

def pairwise_distances_symmetric(Xs, dist, delta = None, num_jobs = 8):
    """ Computes the pairwise edit distances between the objects in
    Xs, assuming that the distance measure is symmetric. Each object in Xs
    needs to be a valid input for the given distance function, i.e. a sequence
    or a tree. Due to symmetry, this method is about double as fast compared
    to pairwise_distances.

    Optionally, it is possible to specify a component-wise distance function
    delta, which will then be forwarded to the input distance function

    Parameters
    ----------
    Xs: list
        a list of sequences or trees.
    dist: function
        a function that takes two elements of Xs as inputs and returns a
        scalar distance value between them.
    delta: function (default = None)
        a function that takes two elements of the input sequences or trees
        as inputs and returns their pairwise distance, where delta(x, None)
        should be the cost of deleting x and delta(None, y) should be the cost
        of inserting y. If this is not None, dist needs to accept an optional
        argument 'delta' as well. Defaults to None.
    num_jobs: int (default = 8)
        The number of jobs to be used for parallel processing. Defaults to 8.

    Returns
    -------
    D: array_like
        a symmetric len(Xs) x len(Xs) matrix of pairwise edit distance values.

    """
    K = len(Xs)
    # set up a parallel processing pool
    pool = mp.Pool(num_jobs)
    # set up the result matrix
    D = np.zeros((K,K))

    # start off all parallel processing jobs.
    # In each job, we compute a _BATCH_SIZE x _BATCH_SIZE patch of the final
    # pairwise distance matrix. Computing batches reduces the overhead of
    # serialization for starting a parallel processing job, because every
    # single worker can do more with the resources it gets.
    results = []
    if(delta is None):
        for k in range(0, K, _BATCH_SIZE):
            k_hi = min(k + _BATCH_SIZE, K)
            for l in range(k, K, _BATCH_SIZE):
                l_hi = min(l + _BATCH_SIZE, K)
                symmetric = l == k
                results.append(pool.apply_async(_batch_dist_with_indices, args=(k, l, dist, Xs[k:k_hi], Xs[l:l_hi], symmetric)))
    else:
        for k in range(0, K, _BATCH_SIZE):
            k_hi = min(k + _BATCH_SIZE, K)
            for l in range(k, K, _BATCH_SIZE):
                l_hi = min(l + _BATCH_SIZE, K)
                symmetric = l == k
                results.append(pool.apply_async(_batch_dist_with_indices_and_delta, args=(k, l, dist, Xs[k:k_hi], Xs[l:l_hi], delta, symmetric)))

    # close the pool
    pool.close()

    # sort all results into the matrix as soon as they arrive
    for res in results:
        k, l, D_batch = res.get()
        k_hi = k + D_batch.shape[0]
        l_hi = l + D_batch.shape[1]
        D[k:k_hi, :][:, l:l_hi] = D_batch

    # ensure that the entire pool is finished
    pool.join()

    # add the lower diagonal
    D += np.transpose(D)

    # return the distance matrix
    return D

def _batch_backtrace_with_indices(k, l, dist, X, Y):
    B = []
    for k2 in range(len(X)):
        b_k2 = []
        for l2 in range(len(Y)):
            b_k2.append(dist(X[k2], Y[l2]))
        B.append(b_k2)
    return (k, l, B)

def _batch_backtrace_with_indices_and_delta(k, l, dist, X, Y, delta):
    B = []
    for k2 in range(len(X)):
        b_k2 = []
        for l2 in range(len(Y)):
            b_k2.append(dist(X[k2], Y[l2], delta))
        B.append(b_k2)
    return (k, l, B)

def pairwise_backtraces(Xs, Ys, dist_backtrace, delta = None, num_jobs = 8):
    """ Computes the pairwise backtraces between the objects in
    Xs and the objects in Ys. Each object in Xs and Ys needs to be a valid
    input for the given distance function, i.e. a sequence or a tree.

    Optionally, it is possible to specify a component-wise distance function
    delta, which will then be forwarded to the input distance function

    Parameters
    ----------
    Xs: list
        a list of sequences or trees.
    Ys: list
        another list of sequences or trees.
    dist_backtrace: function
        a function that takes an element of Xs as first and an
        element of Ys as second input and returns an arbitrary object.
    delta: function (default = None)
        a function that takes two elements of the input sequences or trees
        as inputs and returns their pairwise distance, where
        delta(x, None) should be the cost of deleting x and delta(None, y)
        should be the cost of inserting y. If this is not None, dist needs
        to accept an optional argument 'delta' as well. Defaults to None.
    num_jobs: int (default = 8)
        The number of jobs to be used for parallel processing. Defaults to 8.

    Returns
    -------
    B: list
        a len(Xs) x len(Ys) list of lists of pairwise backtraces.

    """
    K = len(Xs)
    L = len(Ys)
    # set up a parallel processing pool
    pool = mp.Pool(num_jobs)
    # set up the result matrix
    B = []
    for k in range(K):
        bs_k = [None] * L
        B.append(bs_k)

    # start off all parallel processing jobs
    results = []
    if(delta is None):
        for k in range(0, K, _BATCH_SIZE):
            k_hi = min(k + _BATCH_SIZE, K)
            for l in range(0, L, _BATCH_SIZE):
                l_hi = min(l + _BATCH_SIZE, L)
                results.append(pool.apply_async(_batch_backtrace_with_indices, args=(k, l, dist_backtrace, Xs[k:k_hi], Ys[l:l_hi])))
    else:
        for k in range(0, K, _BATCH_SIZE):
            k_hi = min(k + _BATCH_SIZE, K)
            for l in range(0, L, _BATCH_SIZE):
                l_hi = min(l + _BATCH_SIZE, L)
                results.append(pool.apply_async(_batch_backtrace_with_indices_and_delta, args=(k, l, dist_backtrace, Xs[k:k_hi], Ys[l:l_hi], delta)))

    # close the pool
    pool.close()

    # sort all results into the matrix as soon as they arrive
    for res in results:
        k, l, B_batch = res.get()
        for k2 in range(len(B_batch)):
            for l2 in range(len(B_batch[k2])):
                B[k + k2][l + l2] = B_batch[k2][l2]

    # ensure that the entire pool is finished
    pool.join()

    # return the distance matrix
    return B



#!python
#cython: language_level=3
"""
Implements algebraic dynamic programming for edit distances.
"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import heapq
import random
from collections.abc import Callable
import numpy as np
from edist.alignment import Alignment

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

class RuleEntry:
    """ Models the set of all grammar rules for a single nonterminal symbol
    in an ADP grammar.

    Attributes
    ----------
    _reps: list
        A list of possible replacements, stored as tuples of the form
       (rep, B), where rep is the name of a replacement operation and
       B is the nonterminal we obtain after applying rep.
    _dels: list
        A list of possible deletions, stored as tuples of the form
       (del, B), where del is the name of a deletion operation and
       B is the nonterminal we obtain after applying del.
    _inss: list
        A list of possible insertions, stored as tuples of the form
       (ins, B), where ins is the name of a insertion operation and
       B is the nonterminal we obtain after applying ins.

    """
    def __init__(self):
        self._reps = []
        self._dels = []
        self._inss = []

    def __repr__(self):
        op_str = 'replacements: '
        for rep_edge in self._reps:
            op_str += 'via ' + rep_edge[0] + ' to ' + rep_edge[1] + ' '
        op_str += '\ndeletions: '
        for del_edge in self._dels:
            op_str += 'via ' + del_edge[0] + ' to ' + del_edge[1] + ' '
        op_str += '\ninsertions: '
        for ins_edge in self._inss:
            op_str += 'via ' + ins_edge[0] + ' to ' + ins_edge[1] + ' '
        return op_str

    def __str__(self):
        return self.__repr__()


def string_to_index_map(lst):
    """ Inverts a list of objects, i.e. converts a list of objects to
    a map from objects to indices in the list.

    Parameters
    ----------
    lst: list
        A list of objects [`x_1`, ..., `x_m`].

    Returns
    -------
    dct: dictionary
        A map where dct[x_i] = i for all i.

    """
    dct = {}
    for i in range(len(lst)):
        dct[lst[i]] = i
    return dct

def string_to_index_list(lst, dct):
    """ Converts a list of objects to an index list, given a index mapping.

    Parameters
    ----------
    lst: list
        A list of objects [`x_1`, ..., `x_m`].
    dct: dictionary
        A mapping from objects to indices.

    Returns
    -------
    idx_list: list
        The list [dct[`x_1`], ..., dct[`x_m`]]

    """
    idx_list = []
    for e in lst:
        if(e not in dct):
            raise ValueError('unknown entry: %s' % str(e))
        idx_list.append(dct[e])
    return idx_list

def string_to_index_tuple_list(lst, op_dct, nont_dct):
    """ Converts a list of tuples to a list of tuple-indices.

    Parameters
    ----------
    lst: list
        A list of tuples [(`x_1`, `y_1`), ..., (`x_m`, `y_m`)].
    op_dct: dictionary
        A mapping from x-objects to indices.
    nont_dct: dictionary
        A mapping from y-objects to indices.

    Returns
    -------
    idx_list: list
        The list [(op_dct[`x_1`], nont_dct[`y_1`]), ...,
        (op_dct[`x_m`], nont_dct[`y_m`])]

    """
    idx_list = []
    for e in lst:
        op_idx = op_dct[e[0]]
        if(op_idx is None):
            raise ValueError('unknown operation string: %s' % str(e[0]))
        nont_idx = nont_dct[e[1]]
        if(nont_idx is None):
            raise ValueError('unknown nonterminal string: %s' % str(e[1]))
        idx_list.append((op_idx, nont_idx))
    return idx_list

class Grammar:
    """ Models an ADP grammar, consisting of a starting nonterminal, a set of
    accepting nonterminals, a set of permitted edit operations and a set of
    rules of the form

    `A` -> `delta` `B`

    where `A` and `B` are nonterminals and `delta` is an edit operation, either
    a replacement, a deletion, or an insertion. We define the set of possible
    edit scripts permitted by a given grammar inductively as all scripts that
    can be produced by starting from the self._start, replacing the current
    nonterminal with the right-hand-side of a matching rule arbitrary many
    times, and then deleting the current nonterminal if it is accepting.

    Attributes
    ----------
    _nonterminals: list (default = _accepting + {start})
        A list of nonterminals of this grammar. Defaults to the union of
        accepting and start.
    _start: str
        The starting nonterminal, which should be in _nonterminals.
    _accepting: list
        A list of accepting nonterminals, all of which should be in
        self._nonterminals.
    _reps: list (default = [])
        A list of names of possible replacement operations.
        Defaults to an empty list.
    _dels: list (default = [])
        A list of names of possible deletion operations.
        Defaults to an empty list.
    _inss: list (default = [])
        A list of names of possible insertion operations.
        Defaults to an empty list.
    _rules: dictionary (default = {})
        A mapping of nonterminal symbols to RuleEntries. Refer to
        The documentation above for more details on RuleEntries.
        Defaults to an empty map.

    """
    def __init__(self, start, accepting, nonterminals = None, reps = None, dels = None, inss = None, rules = None):
        self._start = start
        self._accepting = accepting
        if(nonterminals is None):
            self._nonterminals = []
        else:
            self._nonterminals = nonterminals
        if(reps is None):
            self._reps = []
        else:
            self._reps = reps
        if(dels is None):
            self._dels = []
        else:
            self._dels = dels
        if(inss is None):
            self._inss = []
        else:
            self._inss = inss
        if(rules is None):
            self._rules = {}
        else:
            self._rules = rules
        self._initialize_rule_entry(start)
        for nont in accepting:
            self._initialize_rule_entry(nont)

    def __repr__(self):
        op_str = 'Start at ' + str(self._start) + '. Rules:\n'
        for nont in self._nonterminals:
            op_str += 'From ' + nont 
            if(nont in self._accepting):
                op_str += ' (accepting)'
            op_str += ':\n'
            op_str += str(self._rules[nont])
        return op_str

    def __str__(self):
        return self.__repr__()

    def _initialize_rule_entry(self, source):
        if(source not in self._nonterminals):
            self._nonterminals.append(source)
        if(source not in self._rules):
            self._rules[source] = RuleEntry()

    def append_replacement(self, source, target, operation):
        """ Appends a rule to this grammar for a replacement operation,
        i.e. a rule of the form `A` -> `rep` `B`.

        Parameters
        ----------
        source: str
            The left-hand-side nonterminal `A` of the new rule.
            If not in self._nonterminals yet, it is appended automatically.
        target: str
            The right-hand-side nonterminal `B` of the new rule.
            If not in self._nonterminals yet, it is appended automatically.
        operation: str
            The name of the replacement operation rep. If not in
            self._reps yet, it is appended automatically.

        """
        if(operation not in self._reps):
            self._reps.append(operation)
        self._initialize_rule_entry(source)
        self._initialize_rule_entry(target)
        self._rules[source]._reps.append((operation, target))

    def append_deletion(self, source, target, operation):
        """ Appends a rule to this grammar for a deletion operation,
        i.e. a rule of the form `A` -> `del` `B`.

        Parameters
        ----------
        source: str
            The left-hand-side nonterminal `A` of the new rule.
            If not in self._nonterminals yet, it is appended automatically.
        target: str
            The right-hand-side nonterminal `B` of the new rule.
            If not in self._nonterminals yet, it is appended automatically.
        operation: str
            The name of the deletion operation del. If not in
            self._dels yet, it is appended automatically.
        """
        if(operation not in self._dels):
            self._dels.append(operation)
        self._initialize_rule_entry(source)
        self._initialize_rule_entry(target)
        self._rules[source]._dels.append((operation, target))

    def append_insertion(self, source, target, operation):
        """ Appends a rule to this grammar for a deletion operation,
        i.e. a rule of the form `A` -> `ins` `B`.

        Parameters
        ----------
        source: str
            The left-hand-side nonterminal `A` of the new rule.
            If not in self._nonterminals yet, it is appended automatically.
        target: str
            The right-hand-side nonterminal `B` of the new rule.
            If not in self._nonterminals yet, it is appended automatically.
        operation: str
            The name of the insertion operation ins. If not in
            self._inss yet, it is appended automatically.

        """
        if(operation not in self._inss):
            self._inss.append(operation)
        self._initialize_rule_entry(source)
        self._initialize_rule_entry(target)
        self._rules[source]._inss.append((operation, target))

    def size(self):
        """ Returns the number of nonterminals in this grammar.

        Returns
        -------
        size: int
            the number of nonterminals in this grammar.

        """
        return len(self._nonterminals)

    def start(self):
        """ Returns the starting nonterminal of this grammar.

        Returns
        -------
        start: str
            the starting nonterminal of this grammar.

        """
        return self._start

    def nonterminals(self):
        """ Returns the list of nonterminals of this grammar.

        Returns
        -------
        nonts: list
            the list of nonterminals of this grammar.

        """
        return self._nonterminals

    def validate(self, deltas):
        """ Ensures that this grammar is compatible with the given algebra
        deltas.

        Parameters
        ----------
        deltas: dictionary
            An algebra, i.e. a mapping from operation names to distance
            functions.

        Raises
        ------
        ValueError
            If any of the operations of this grammar is not supported by the
            given algebra.

        """
        for rep_op in self._reps:
            if(rep_op not in deltas):
                raise ValueError('costs for the ' + rep_op + ' operation are undefined!')
        for del_op in self._dels:
            if(del_op not in deltas):
                raise ValueError('costs for the ' + del_op + ' operation are undefined!')
        for ins_op in self._inss:
            if(ins_op not in deltas):
                raise ValueError('costs for the ' + ins_op + ' operation are undefined!')

    def adjacency_lists(self):
        """ Returns the adjacency list format of this grammar.

        Returns
        -------
        start_idx: int
            The index of the starting nonterminal symbol.
        accpt_idxs: list
            A list of indices for all accepting nonterminals.
        rep_adj: list
            An adjacency list covering all replacement operations,
            i.e. a list where the `i`-th entry contains all rules of the form
            `A` -> `rep` `B`, where `A` is the `i`-th nonterminal and the
            right-hand-sides are represented as tuples of operation indices and
            nonterminal indices.
        del_adj: list
            An adjacency list covering all deletion operations,
            i.e. a list where the `i`-th entry contains all rules of the form
            `A` -> `del` `B`, where `A` is the `i`-th nonterminal and the
            right-hand-sides are represented as tuples of operation indices and
            nonterminal indices.
        ins_adj: list
            An adjacency list covering all insertion operations,
            i.e. a list where the `i`-th entry contains all rules of the form
            `A` -> `ins` `B`, where `A` is the `i`-th nonterminal and the
            right-hand-sides are represented as tuples of operation indices and
            nonterminal indices.

        """
        # first, create maps from string to index representations
        nont_map = string_to_index_map(self._nonterminals)
        reps_map = string_to_index_map(self._reps)
        dels_map = string_to_index_map(self._dels)
        inss_map = string_to_index_map(self._inss)
        # translate the start symbol to an index
        start_idx = nont_map[self._start]
        # translate the accepting symbol list
        accpt_idxs = string_to_index_list(self._accepting, nont_map)
        # then, create an adjacency list representation for all replacements,
        # deletions, and insertions separately
        rep_adj = []
        del_adj = []
        ins_adj = []
        for nont in self._nonterminals:
            rule_entry = self._rules[nont]
            rep_adj.append(string_to_index_tuple_list(rule_entry._reps, reps_map, nont_map))
            del_adj.append(string_to_index_tuple_list(rule_entry._dels, dels_map, nont_map))
            ins_adj.append(string_to_index_tuple_list(rule_entry._inss, inss_map, nont_map))
        return start_idx, accpt_idxs, rep_adj, del_adj, ins_adj

    def inverse_adjacency_lists(self):
        """ Returns the inverse adjacency list format of this grammar.

        Returns
        -------
        start_idx: int
            The index of the starting nonterminal symbol.
        accpt_idxs: list
            A list of indices for all accepting nonterminals.
        rep_adj: list
            An adjacency list covering all replacement operations,
            i.e. a list where the `i`-th entry contains all rules of the form
            `A` -> `rep` `B`, where `B` is the `i`-th nonterminal and
            (`A`, `rep`) is represented as a tuple tuples of operation index
            and nonterminal index.
        del_adj: list
            An adjacency list covering all deletion operations,
            i.e. a list where the `i`-th entry contains all rules of the form
            `A` -> `del` `B`, where `B` is the `i`-th nonterminal and
            (`A`, `del`) is represented as a tuple tuples of operation index
            and nonterminal index.
        ins_adj: list
            An adjacency list covering all insertion operations,
            i.e. a list where the `i`-th entry contains all rules of the form
            `A` -> `ins` `B`, where `B` is the `i`-th nonterminal and
            (`A`, `ins`) is represented as a tuple tuples of operation index
            and nonterminal index.

        """
        # first, create maps from string to index representations
        nont_map = string_to_index_map(self._nonterminals)
        reps_map = string_to_index_map(self._reps)
        dels_map = string_to_index_map(self._dels)
        inss_map = string_to_index_map(self._inss)
        # translate the start symbol to an index
        start_idx = nont_map[self._start]
        # translate the accepting symbol list
        accpt_idxs = string_to_index_list(self._accepting, nont_map)
        # then, create an adjacency list representation for all replacements,
        # deletions, and insertions separately
        rep_adj = []
        del_adj = []
        ins_adj = []
        for nont in self._nonterminals:
            rep_adj.append([])
            del_adj.append([])
            ins_adj.append([])
        for A in self._nonterminals:
            rule_entry = self._rules[A]
            for (delta, B) in rule_entry._reps:
                rep_adj[nont_map[B]].append((reps_map[delta], nont_map[A]))
            for (delta, B) in rule_entry._dels:
                del_adj[nont_map[B]].append((dels_map[delta], nont_map[A]))
            for (delta, B) in rule_entry._inss:
                ins_adj[nont_map[B]].append((inss_map[delta], nont_map[A]))
        return start_idx, accpt_idxs, rep_adj, del_adj, ins_adj

def edit_distance(x, y, grammar, deltas):
    """ Computes the edit distance between two sequences x and y, based on
    the given ADP grammar and the given algebra.

    Parameters
    ----------
    x: list
        A list of objects.
    y: list
        Another list of objects.
    grammar: class adp.Grammar
        An ADP grammar. Refer to the documentation above for more information.
    deltas: dictionary
        An algebra, i.e. a mapping from operation names to distance functions
        OR a single distance function if the grammar supports only a single
        replacement, deletion, and insertion operation.

    Returns
    -------
    d: float
        The edit distance between x and y.

    """
    # apply the internal edit distance function
    Ds, _, _, _, start_idx, _, _, _, _ = _edit_distance(x, y, grammar, deltas)
    return Ds[start_idx, 0, 0]

def _edit_distance(x, y, grammar, deltas):
    """ Computes the edit distance including all internal variables
    necessary during computation.

    Parameters
    ----------
    x: list
        A list of objects of length m.
    y: list
        Another list of objects of length n.
    grammar: class adp.Grammar
        An ADP grammar with R nonterminals, K_rep replacements,
        K_del deletions, and K_ins insertions. Refer to the documentation
        above for more information.
    deltas: dictionary
        An algebra, i.e. a mapping from operation names to distance functions
        OR a single distance function if the grammar supports only a single
        replacement, deletion, and insertion operation.

    Returns
    -------
    Ds: array_like
        A R x m+1 x n+1 tensor containing the dynamic programming matrices
        for all nonterminals.
    Deltas_rep: array_like
        A K_rep x m x n tensor containing the replacement costs for
        all replacements.
    Deltas_del: array_like
        A K_rep x m matrix containing the deletion costs for
        all deletions.
    Deltas_ins: array_like
        A K_ins x n matrix containing the insertion costs for
        all insertions.
    start_idx: int
        The index of the starting nonterminal.
    accpt_idxs: list
        The indices of accepting nonterminals.
    adj_rep: list
        An adjacency list representation of the grammar rules for
        replacement operations.
    adj_del: list
        An adjacency list representation of the grammar rules for
        deletion operations.
    adj_ins: list
        An adjacency list representation of the grammar rules for
        insertion operations.

    """
    # check if the given algebra is compatible with the given grammar.
    if(isinstance(deltas, Callable)):
        if(len(grammar._reps) > 1 or len(grammar._dels) > 1 or len(grammar._inss) > 1):
            raise ValueError('If a function is given instead of an algebra, the grammar can only support a single operation of each type; otherwise, ambiguities arise.')
        # generate a mock algebra that is definitely compatible
        delta = deltas
        deltas = {grammar._reps[0] : delta, grammar._dels[0] : delta, grammar._inss[0] : delta}
    else:
        grammar.validate(deltas)

    cdef int m = len(x)
    cdef int n = len(y)
    # pre-compute all operation costs

    # First, compute all pairwise replacements
    cdef int K_rep = len(grammar._reps)
    Deltas_rep = np.zeros((K_rep, m, n))
    cdef double[:,:,:] Deltas_rep_view = Deltas_rep
    cdef int i
    cdef int j
    cdef int k
    for k in range(K_rep):
        delta = deltas[grammar._reps[k]]
        for i in range(m):
            for j in range(n):
                Deltas_rep_view[k, i, j] = delta(x[i], y[j])

    # Then, compute all deletions
    cdef int K_del = len(grammar._dels)
    Deltas_del = np.zeros((K_del, m))
    cdef double[:,:] Deltas_del_view = Deltas_del
    for k in range(K_del):
        delta = deltas[grammar._dels[k]]
        for i in range(m):
            Deltas_del_view[k, i] = delta(x[i], None)

    # Then, compute all insertions
    cdef int K_ins = len(grammar._inss)
    Deltas_ins = np.zeros((K_ins, n))
    cdef double[:,:] Deltas_ins_view = Deltas_ins
    for k in range(K_ins):
        delta = deltas[grammar._inss[k]]
        for j in range(n):
            Deltas_ins_view[k, j] = delta(None, y[j])

    # retrieve the adjacency list representation for
    # the grammar
    start_idx, accpt_idxs, adj_rep, adj_del, adj_ins = grammar.adjacency_lists()

    # Initialize the dynamic programming matrices
    # for all nonterminals
    cdef int R = len(grammar._nonterminals)
    Ds = np.full((R, m+1, n+1), np.inf)
    cdef double[:,:,:] Ds_view = Ds
    # initialize last entry for all accepting symbols
    cdef int nont
    for nont in accpt_idxs:
        Ds_view[nont, m, n] = 0.

    # initialize last column for all symbols
    cdef int r
    cdef int s
    for i in range(m-1,-1,-1):
        for r in range(R):
            for (k, s) in adj_del[r]:
                Ds_view[r, i, n] = Deltas_del_view[k, i] + Ds_view[s, i+1, n]

    # initialize last row for all symbols
    for j in range(n-1,-1,-1):
        for r in range(R):
            for (k, s) in adj_ins[r]:
                Ds_view[r, m, j] = Deltas_ins_view[k, j] + Ds_view[s, m, j+1]

    # perform the remaining computation
    cdef double min_cost
    cdef double current_cost
    for i in range(m-1,-1,-1):
        for j in range(n-1,-1,-1):
            for r in range(R):
                min_cost = np.inf
                # first, consider replacements
                for (k, s) in adj_rep[r]:
                    current_cost = Deltas_rep_view[k, i, j] + Ds_view[s, i+1, j+1]
                    if(current_cost < min_cost):
                        min_cost = current_cost
                # then, consider deletions
                for (k, s) in adj_del[r]:
                    current_cost = Deltas_del_view[k, i] + Ds_view[s, i+1, j]
                    if(current_cost < min_cost):
                        min_cost = current_cost
                # finally, consider insertions
                for (k, s) in adj_ins[r]:
                    current_cost = Deltas_ins_view[k, j] + Ds_view[s, i, j+1]
                    if(current_cost < min_cost):
                        min_cost = current_cost
                # set new entry to minimum
                Ds_view[r, i, j] = min_cost

    return Ds, Deltas_rep, Deltas_del, Deltas_ins, start_idx, accpt_idxs, adj_rep, adj_del, adj_ins


####### BACKTRACING FUNCTIONS #######

cdef double _BACKTRACE_TOL = 1E-5

def backtrace(x, y, grammar, deltas):
    """ Computes a co-optimal alignment between the two input sequences
    x and y, given the given ADP grammar and algebra. This mechanism
    is deterministic and will always prefer replacements over other options.

    Parameters
    ----------
    x: list
        A list of objects.
    y: list
        Another list of objects.
    grammar: class adp.Grammar
        An ADP grammar. Refer to the documentation above for more information.
    deltas: dictionary
        An algebra, i.e. a mapping from operation names to distance functions
        OR a single distance function if the grammar supports only a single
        replacement, deletion, and insertion operation.

    Returns
    -------
    alignment: class alignment.Alignment
        a co-optimal alignment between x and y.

    """
    # apply the internal edit distance function
    Ds, Deltas_rep, Deltas_del, Deltas_ins, start_idx, accpt_idxs, adj_rep, adj_del, adj_ins = _edit_distance(x, y, grammar, deltas)

    # declare c variables
    cdef int m = len(x)
    cdef int n = len(y)
    cdef int i
    cdef int j
    cdef int k
    cdef int r
    cdef int s

    cdef double[:,:,:] Ds_view = Ds
    cdef double[:,:, :] Deltas_rep_view = Deltas_rep
    cdef double[:,:] Deltas_del_view = Deltas_del
    cdef double[:,:] Deltas_ins_view = Deltas_ins

    # then, perform the backtracing
    r = start_idx
    i = 0
    j = 0
    cdef int found_coopt
    alignment = Alignment()
    while(i < m and j < n):
        # check which alignment option is co-optimal
        found_coopt = False
        # check all possible replacements
        for (k, s) in adj_rep[r]:
            current_cost = Deltas_rep_view[k, i, j] + Ds_view[s, i+1, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # replacement is co-optimal
                alignment.append_tuple(i, j, grammar._reps[k])
                r = s
                i += 1
                j += 1
                found_coopt = True
                break
        if(found_coopt):
            continue
        # check all possible deletions
        for (k, s) in adj_del[r]:
            current_cost = Deltas_del_view[k, i] + Ds_view[s, i+1, j]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # deletion is co-optimal
                alignment.append_tuple(i, -1, grammar._dels[k])
                r = s
                i += 1
                found_coopt = True
                break
        if(found_coopt):
            continue
        # check all possible insertions
        for (k, s) in adj_ins[r]:
            current_cost = Deltas_ins_view[k, j] + Ds_view[s, i, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # insertion is co-optimal
                alignment.append_tuple(-1, j, grammar._inss[k])
                r = s
                j += 1
                found_coopt = True
                break
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')
    # delete all demaining nodes
    while(i < m):
        found_coopt = False
        # check all possible deletions
        for (k, s) in adj_del[r]:
            current_cost = Deltas_del_view[k, i] + Ds_view[s, i+1, j]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # deletion is co-optimal
                alignment.append_tuple(i, -1, grammar._dels[k])
                r = s
                i += 1
                found_coopt = True
                break
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')
    # insert all demaining nodes
    while(j < n):
        found_coopt = False
        # check all possible deletions
        for (k, s) in adj_ins[r]:
            current_cost = Deltas_ins_view[k, j] + Ds_view[s, i, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # insertion is co-optimal
                alignment.append_tuple(-1, j, grammar._inss[k])
                r = s
                j += 1
                found_coopt = True
                break
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')
    if(r not in accpt_idxs):
        raise ValueError('Ended in a non-accepting nonterminal')
    return alignment

def backtrace_stochastic(x, y, grammar, deltas):
    """ Computes a co-optimal alignment between the two input sequences
    x and y, given the given ADP grammar and algebra. This mechanism
    is stochastic and will return a random alignment.

    Note that the randomness does _not_ produce a uniform distribution over
    all co-optimal alignments because random choices at the start of the
    alignment process dominate. If you wish to characterize the overall
    distribution accurately, use backtrace_matrix instead.

    Parameters
    ----------
    x: list
        A list of objects.
    y: list
        Another list of objects.
    grammar: class adp.Grammar
        An ADP grammar. Refer to the documentation above for more information.
    deltas: dictionary
        An algebra, i.e. a mapping from operation names to distance functions
        OR a single distance function if the grammar supports only a single
        replacement, deletion, and insertion operation.

    Returns
    -------
    alignment: class alignment.Alignment
        a co-optimal alignment between x and y.

    """
    # apply the internal edit distance function
    Ds, Deltas_rep, Deltas_del, Deltas_ins, start_idx, accpt_idxs, adj_rep, adj_del, adj_ins = _edit_distance(x, y, grammar, deltas)

    # declare c variables
    cdef int m = len(x)
    cdef int n = len(y)
    cdef int i
    cdef int j
    cdef int k
    cdef int r
    cdef int s
    cdef int c
    cdef int op

    cdef double[:,:,:] Ds_view = Ds
    cdef double[:,:,:] Deltas_rep_view = Deltas_rep
    cdef double[:,:] Deltas_del_view = Deltas_del
    cdef double[:,:] Deltas_ins_view = Deltas_ins

    # then, perform the backtracing
    r = start_idx
    i = 0
    j = 0
    cdef int found_coopt
    alignment = Alignment()
    while(i < m and j < n):
        # check which alignment option is co-optimal
        # check all possible replacements
        coopts = []
        for (k, s) in adj_rep[r]:
            current_cost = Deltas_rep_view[k, i, j] + Ds_view[s, i+1, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # replacement is co-optimal
                coopts.append((0, k, s))
        # check all possible deletions
        for (k, s) in adj_del[r]:
            current_cost = Deltas_del_view[k, i] + Ds_view[s, i+1, j]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # deletion is co-optimal
                coopts.append((1, k, s))
        # check all possible insertions
        for (k, s) in adj_ins[r]:
            current_cost = Deltas_ins_view[k, j] + Ds_view[s, i, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # insertion is co-optimal
                coopts.append((2, k, s))
        if(not coopts):
            raise ValueError('Internal error: No option is co-optimal.')
        # select a random operation
        c = random.randrange(len(coopts))
        # apply the corresponding operation
        op = coopts[c][0]
        k  = coopts[c][1]
        r  = coopts[c][2]
        if(op == 0):
            # apply replacement
            alignment.append_tuple(i, j, grammar._reps[k])
            i += 1
            j += 1
        elif(op == 1):
            # apply deletion
            alignment.append_tuple(i, -1, grammar._dels[k])
            i += 1
        elif(op == 2):
            # apply insertion
            alignment.append_tuple(-1, j, grammar._inss[k])
            j += 1
    # delete all demaining nodes
    while(i < m):
        coopts = []
        # check all possible deletions
        for (k, s) in adj_del[r]:
            current_cost = Deltas_del_view[k, i] + Ds_view[s, i+1, j]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # deletion is co-optimal
                coopts.append((k, s))
        if(not coopts):
            raise ValueError('Internal error: No option is co-optimal.')
        # select a random deletion
        c = random.randrange(len(coopts))
        # apply the corresponding operation
        k  = coopts[c][0]
        r  = coopts[c][1]
        alignment.append_tuple(i, -1, grammar._dels[k])
        i += 1

    # insert all demaining nodes
    while(j < n):
        coopts = []
        # check all possible insertions
        for (k, s) in adj_ins[r]:
            current_cost = Deltas_ins_view[k, j] + Ds_view[s, i, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # insertion is co-optimal
                coopts.append((k, s))
        if(not coopts):
            raise ValueError('Internal error: No option is co-optimal.')
        # select a random insertion
        c = random.randrange(len(coopts))
        # apply the corresponding operation
        k  = coopts[c][0]
        r  = coopts[c][1]
        alignment.append_tuple(-1, j, grammar._inss[k])
        j += 1
    if(r not in accpt_idxs):
        raise ValueError('Ended in a non-accepting nonterminal')
    return alignment


def backtrace_matrix(x, y, grammar, deltas):
    """ Computes three tensors, P_rep, P_del, and P_ins, which summarize all
    co-optimal alignments between x and y.

    In particular, P_rep[k, i, j] specifies the fraction of co-optimal
    alignments in which node x[i] has been replaced with node y[j] using
    operation grammar._reps[k]. Accordingly, P_del[k, i] specifies the
    fraction of co-optimal alignments in which x[i] has been deleted using
    operation grammar._dels[k], and P_ins[k, j] specifies the fraction of
    co-optimal alignments in which y[j] has been inserted using operation
    grammar._inss[k].

    Parameters
    ----------
    x: list
        A list of objects.
    y: list
        Another list of objects.
    grammar: class adp.Grammar
        An ADP grammar. Refer to the documentation above for more information.
    deltas: dictionary
        An algebra, i.e. a mapping from operation names to distance functions
        OR a single distance function if the grammar supports only a single
        replacement, deletion, and insertion operation.

    Returns
    -------
    P_rep: array_like
        a tensor where P_rep[k, i, j] specifies the fraction of co-optimal
        alignments in which node x[i] has been replaced with node y[j] using
        operation grammar._reps[k].
    P_del: array_like
        a matrix where P_del[k, i] specifies the fraction of co-optimal
        alignments in which x[i] has been deleted using operation
        grammar._dels[k].
    P_ins: array_like
        a matrix where P_ins[k, j] specifies the fraction of co-optimal
        alignments in which y[j] has been inserted using operation
        grammar._inss[k].
    k: int
        the number of co-optimal alignments.

    """
    # apply the internal edit distance function
    Ds, Deltas_rep, Deltas_del, Deltas_ins, start_idx, accpt_idxs, adj_rep, adj_del, adj_ins = _edit_distance(x, y, grammar, deltas)

    # declare c variables
    cdef int m = len(x)
    cdef int n = len(y)
    cdef int i
    cdef int j
    cdef int k
    cdef int r
    cdef int s
    cdef int c
    cdef int op

    cdef double[:,:,:] Ds_view = Ds
    cdef double[:,:,:] Deltas_rep_view = Deltas_rep
    cdef double[:,:] Deltas_del_view = Deltas_del
    cdef double[:,:] Deltas_ins_view = Deltas_ins

    # compute the forward tensor Alpha, which contains the number of
    # co-optimal alignment paths from cell [start_idx, 0, 0] to cell [r, i, j]
    Alpha = np.zeros((len(grammar._nonterminals), m+1, n+1), dtype=int)
    cdef long[:,:, :] Alpha_view = Alpha
    Alpha_view[start_idx, 0, 0] = 1
    # build a queue of cells which we still need to process
    q = [(start_idx, 0, 0)]
    # build a set which stores the already visited cells
    visited = set()
    # initialize temporary variables
    cdef int found_coopt = False
    cdef long num_coopts = 0
    while(q):
        (r, i, j) = heapq.heappop(q)
        if((i, j) in visited):
            continue
        visited.add((r, i, j))
        num_coopts = Alpha_view[r, i, j]
        found_coopt = False
        if(i == m):
            if(j == n):
                continue
            # if we are at the end of the first sequence, we can only insert
            # check all possible insertions
            for (k, s) in adj_ins[r]:
                current_cost = Deltas_ins_view[k, j] + Ds_view[s, i, j+1]
                if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                    # insertion is co-optimal
                    Alpha_view[s, i, j+1] += num_coopts
                    heapq.heappush(q, (s, i, j+1))
                    found_coopt = True
            if(not found_coopt):
                raise ValueError('Internal error: No option is co-optimal.')
            continue
        if(j == n):
            # if we are at the end of the second sequence, we can only delete
            # check all possible deletions
            for (k, s) in adj_del[r]:
                current_cost = Deltas_del_view[k, i] + Ds_view[s, i+1, j]
                if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                    # deletion is co-optimal
                    Alpha_view[s, i+1, j] += num_coopts
                    heapq.heappush(q, (s, i+1, j))
                    found_coopt = True
            if(not found_coopt):
                raise ValueError('Internal error: No option is co-optimal.')
            continue
        # check which alignment option is co-optimal
        # check all possible replacements
        for (k, s) in adj_rep[r]:
            current_cost = Deltas_rep_view[k, i, j] + Ds_view[s, i+1, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # replacement is co-optimal
                Alpha_view[s, i+1, j+1] += num_coopts
                heapq.heappush(q, (s, i+1, j+1))
                found_coopt = True
        # check all possible deletions
        for (k, s) in adj_del[r]:
            current_cost = Deltas_del_view[k, i] + Ds_view[s, i+1, j]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # deletion is co-optimal
                Alpha_view[s, i+1, j] += num_coopts
                heapq.heappush(q, (s, i+1, j))
                found_coopt = True
        # check all possible insertions
        for (k, s) in adj_ins[r]:
            current_cost = Deltas_ins_view[k, j] + Ds_view[s, i, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # insertion is co-optimal
                Alpha_view[s, i, j+1] += num_coopts
                heapq.heappush(q, (s, i, j+1))
                found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    # compute the backward matrix Beta, which contains the number of
    # co-optimal alignment paths from cell [s, i, j] to cells [accepting, m, n]
    Beta = np.zeros((len(grammar._nonterminals), m+1, n+1), dtype=int)
    cdef long[:,:,:] Beta_view = Beta
    for r in accpt_idxs:
        Beta_view[r, m, n] = 1
    # retrieve the inverse adjacency list format of the grammar
    _, _, inv_adj_rep, inv_adj_del, inv_adj_ins = grammar.inverse_adjacency_lists()

    # iterate in downward lexigraphic order over the visited cells
    for (s, i, j) in sorted(visited, key = (lambda tpl : (tpl[1], tpl[2], tpl[0])), reverse = True):
        num_coopts = Beta_view[s, i, j]
        found_coopt = False
        if(i == 0):
            if(j == 0):
                continue
            # if we are at the start of the first sequence, we can only insert
            # check all possible insertions
            for (k, r) in inv_adj_ins[s]:
                current_cost = Deltas_ins_view[k, j-1] + Ds_view[s, i, j]
                if(Ds_view[r, i, j-1] + _BACKTRACE_TOL > current_cost):
                    # insertion is co-optimal
                    Beta_view[r, i, j-1] += num_coopts
                    found_coopt = True
            if(not found_coopt):
                raise ValueError('Internal error: No option is co-optimal.')
            continue
        if(j == 0):
            # if we are at the start of the second sequence, we can only delete
            # check all possible deletions
            for (k, r) in inv_adj_del[s]:
                current_cost = Deltas_del_view[k, i-1] + Ds_view[s, i, j]
                if(Ds_view[r, i-1, j] + _BACKTRACE_TOL > current_cost):
                    # insertion is co-optimal
                    Beta_view[r, i-1, j] += num_coopts
                    found_coopt = True
            if(not found_coopt):
                raise ValueError('Internal error: No option is co-optimal.')
            continue
        # check which alignment option is co-optimal
        # check all possible replacements
        for (k, r) in inv_adj_rep[s]:
            current_cost = Deltas_rep_view[k, i-1, j-1] + Ds_view[s, i, j]
            if(Ds_view[r, i-1, j-1] + _BACKTRACE_TOL > current_cost):
                # insertion is co-optimal
                Beta_view[r, i-1, j-1] += num_coopts
                found_coopt = True
        # check all possible deletions
        for (k, r) in inv_adj_del[s]:
            current_cost = Deltas_del_view[k, i-1] + Ds_view[s, i, j]
            if(Ds_view[r, i-1, j] + _BACKTRACE_TOL > current_cost):
                # insertion is co-optimal
                Beta_view[r, i-1, j] += num_coopts
                found_coopt = True
        # check all possible insertions
        for (k, r) in inv_adj_ins[s]:
            current_cost = Deltas_ins_view[k, j-1] + Ds_view[s, i, j]
            if(Ds_view[r, i, j-1] + _BACKTRACE_TOL > current_cost):
                # insertion is co-optimal
                Beta_view[r, i, j-1] += num_coopts
                found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    if(np.sum(Alpha[:, m, n]) != Beta_view[start_idx, 0, 0]):
        raise ValueError('Internal error: Alignment count in Alpha and Beta matrix did not agree; got %d versus %d' % (np.sum(Alpha[:, m, n]), Beta_view[start_idx, 0, 0]))

    # compute the output matrices, specifying how often each operation was used
    K_rep = np.zeros((len(grammar._reps), m, n), dtype=int)
    cdef long[:,:,:] K_rep_view = K_rep
    K_del = np.zeros((len(grammar._dels), m), dtype=int)
    cdef long[:,:] K_del_view = K_del
    K_ins = np.zeros((len(grammar._inss), n), dtype=int)
    cdef long[:,:] K_ins_view = K_ins
    for (r, i, j) in visited:
        num_coopts = Alpha_view[r, i, j]
        if(i == m):
            if(j == n):
                continue
            # if we are at the end of the first sequence, we can only insert
            # check all possible insertions
            for (k, s) in adj_ins[r]:
                current_cost = Deltas_ins_view[k, j] + Ds_view[s, i, j+1]
                if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                    # insertion is co-optimal
                    K_ins_view[k, j] += num_coopts * Beta_view[s, i, j+1]
            continue
        if(j == n):
            # if we are at the end of the second sequence, we can only delete
            # check all possible deletions
            for (k, s) in adj_del[r]:
                current_cost = Deltas_del_view[k, i] + Ds_view[s, i+1, j]
                if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                    # deletion is co-optimal
                    K_del_view[k, i] += num_coopts * Beta_view[s, i+1, j]
            continue
        # check which alignment option is co-optimal
        # check all possible replacements
        for (k, s) in adj_rep[r]:
            current_cost = Deltas_rep_view[k, i, j] + Ds_view[s, i+1, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # replacement is co-optimal
                K_rep_view[k, i, j] += num_coopts * Beta_view[s, i+1, j+1]
        # check all possible deletions
        for (k, s) in adj_del[r]:
            current_cost = Deltas_del_view[k, i] + Ds_view[s, i+1, j]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # deletion is co-optimal
                K_del_view[k, i] += num_coopts * Beta_view[s, i+1, j]
        # check all possible insertions
        for (k, s) in adj_ins[r]:
            current_cost = Deltas_ins_view[k, j] + Ds_view[s, i, j+1]
            if(Ds_view[r, i, j] + _BACKTRACE_TOL > current_cost):
                # insertion is co-optimal
                K_ins_view[k, j] += num_coopts * Beta_view[s, i, j+1]

    # compute the final summary matrices by dividing K by the overall number
    # of co-optimal alignments
    num_coopts = Beta_view[start_idx, 0, 0]
    P_rep = K_rep.astype(float) / num_coopts
    P_del = K_del.astype(float) / num_coopts
    P_ins = K_ins.astype(float) / num_coopts
    return P_rep, P_del, P_ins, num_coopts



"""
Implements an alignment between two sequences or trees.

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

class Tuple:
    """ Models a single alignment entry with an edit operation name,
    a left index, and a right index.

    Attributes
    ----------
    _name: str
        The name of the corresponding edit operation.
    _left: int
        The index of the aligned object on the left or -1 if no such object
        exists.
    _right: int
        The index of the aligned object on the right or -1 if no such object
        exists.

    """
    def __init__(self, name, left, right):
        self._name = name
        self._left = left
        self._right = right

    def cost(self, x, y, deltas):
        """ Computes the cost of the current edit tuple.

        Parameters
        ----------
        x: list
            A symbol list for the left indices.
        y: list
            A symbol list for the right indices.
        deltas: function or dictionary
            The cost function delta mapping pairs of elements to
            replacement/deletion/insertion costs OR
            A map which contains for any operation name such a function.

        Returns
        -------
        cost: float
            The cost assigned by deltas to this tuple.

        """
        if(self._left >= 0):
            left = x[self._left]
        else:
            left = None
        if(self._right >= 0):
            right = y[self._right]
        else:
            right = None
        if(self._name):
            delta = deltas[self._name]
            return delta(left, right)
        else:
            return deltas(left, right)


    def render(self, x, y, deltas = None):
        """ Represents an tuple as a string, showing the left
        and right indices in addition to the respective labels in x and y,
        and in addition to the tuple cost.

        Parameters
        ----------
        x: list
            A symbol list for the left indices.
        y: list
            A symbol list for the right indices.
        deltas: function or dictionary (default = None)
            The cost function delta mapping pairs of elements to
            replacement/deletion/insertion costs OR
            A map which contains for any operation name such a function.
            If provided, the cost for any operation is rendered as well.

        Returns
        -------
        repr: str
            A string representing this tuple.

        """
        op_str = ''
        if(self._name):
            op_str += str(self._name)
            op_str += ': '
        if(self._left >= 0):
            left = x[self._left]
            op_str += str(left) + ' [%d]' % self._left
        else:
            left = None
            op_str += '-'
        op_str += ' vs. '
        if(self._right >= 0):
            right = y[self._right]
            op_str += str(right) + ' [%d]' % self._right
        else:
            right = None
            op_str += '-'
        if(deltas):
            op_str += ': '
            if(self._name):
                delta = deltas[self._name]
                op_str += str(delta(left, right))
            else:
                op_str += str(deltas(left, right))
        return op_str

    def __repr__(self):
        op_str = ''
        if(self._name):
            op_str += str(self._name)
            op_str += ': '
        if(self._left >= 0):
            op_str += str(self._left)
        else:
            op_str += '-'
        op_str += ' vs. '
        if(self._right >= 0):
            op_str += str(self._right)
        else:
            op_str += '-'
        return op_str

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return isinstance(other, Tuple) and self._name == other._name and self._left == other._left and self._right == other._right

class Alignment(list):
    """ Models a list of tuples. Note that, by convention, an alignment between
    sequences only permits tuples in lexicographically ascending order, i.e.
    an alignment of nothing to 0, 0 to 1, and 1 to 2, should be stored in that
    order and not as [(1, 2), (0, 1), (-1, 0)], for example. The same holds for
    tree alignments, with the additional requirement that aligned indices must
    respect the structure of the tree, i.e. if i is aligned to j and i2 to j2,
    then i can only be a parent of i2 if j is a parent of j2 (and vice versa).

    """
    def __init__(self):
        list.__init__(self, [])

    def append_tuple(self, left, right, op = None):
        """ Appends a new tuple to the current Alignment.

        Parameters
        ----------
        left: int
            the left index.
        right: int
            the right index.
        op: str (default = None)
            a name for the underlying edit operation.

        """
        self.append(Tuple(op, left, right))

    def cost(self, x, y, deltas):
        """ Computes the cost of this trace. This is equivalent to
        the sum of the cost of all tuples in this trace.

        Parameters
        ----------
        x: list
            A symbol list for the left indices.
        y: list
            A symbol list for the right indices.
        deltas: function or dictionary
            The cost function delta mapping pairs of elements to
            replacement/deletion/insertion costs OR
            A map which contains for any operation name such a function.

        Returns
        -------
        cost: float
            The cost assigned by deltas to this Alignment.

        """
        d = 0.
        for op in self:
            d += op.cost(x, y, deltas)
        return d

    def render(self, x, y, deltas = None):
        """ Represents this trace as a string, showing the left
        and right indices in addition to the respective labels in x and y,
        and in addition to the tuple cost. This is equivalent as to
        calling 'render' on all tuples in this trace and joining the
        resulting strings with newlines.

        Parameters
        ----------
        x: list
            A symbol list for the left indices.
        y: list
            A symbol list for the right indices.
        deltas: function or dictionary (default = None)
            The cost function delta mapping pairs of elements to
            replacement/deletion/insertion costs OR
            A map which contains for any operation name such a function.
            If provided, the cost for any operation is rendered as well.

        Returns
        -------
        repr: str
            A string representing this Alignment.

        """
        render =  []
        for op in self:
            render.append(op.render(x, y, deltas))
        return '\n'.join(render)



#!python
#cython: language_level=3
"""
Implements the dynamic time warping distance of Vintsyuk (1968) and its
backtracing in cython.

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import random
import heapq
import numpy as np
from cython.parallel import prange
from libc.math cimport sqrt
cimport cython
from edist.alignment import Alignment

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

def dtw(x, y, delta):
    """ Computes the dynamic time warping distance between the input sequence
    x and the input sequence y, given the element-wise distance function delta.

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.
    delta: function
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them.

    Returns
    -------
    d: float
        the dynamic time warping distance between x and y according to delta.

    """
    cdef int m = len(x)
    cdef int n = len(y)
    if(m < 1 or n < 1):
        raise ValueError('Dynamic time warping can not handle empty input sequences!')
    # First, compute all pairwise replacements
    Delta = np.zeros((m, n))
    cdef double[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            Delta_view[i,j] = delta(x[i], y[j])

    # Then, compute the dynamic time warping distance
    D = np.zeros((m,n))
    dtw_c(Delta, D)
    return D[0,0]

@cython.boundscheck(False)
def dtw_numeric(double[:] x, double[:] y):
    """ Computes the dynamic time warping distance between two input arrays x
    and y, using the absolute value as element-wise distance measure.

    Parameters
    ----------
    x: array_like
        an array of doubles.
    y: array_like
        another array of doubles.

    Returns
    -------
    d: float
        the dynamic time warping distance between x and y.

    """
    cdef int m = len(x)
    cdef int n = len(y)
    if(m < 1 or n < 1):
        raise ValueError('Dynamic time warping can not handle empty input sequences!')
    # First, compute all pairwise replacements
    # using OMP parallelization
    cdef int i
    cdef int j
    Delta = np.zeros((m, n))
    cdef double[:,:] Delta_view = Delta
    for i in prange(m, nogil=True):
        for j in prange(n):
            if(x[i] > y[j]):
                Delta_view[i,j] = x[i] - y[j]
            else:
                Delta_view[i,j] = y[j] - x[i]
    # Then, compute the dynamic time warping distance
    D = np.zeros((m,n))
    dtw_c(Delta, D)
    return D[0,0]

@cython.boundscheck(False)
def dtw_manhattan(double[:,:] x, double[:,:] y):
    """ Computes the multivariate dynamic time warping distance between two
    input arrays x and y, using the Manhattan distance as element-wise
    distance measure.

    Parameters
    ----------
    x: array_like
        a m x K matrix of doubles.
    y: array_like
        a n x K matrix of doubles.

    Returns
    -------
    d: float
        the dynamic time warping distance between x and y.

    """
    cdef int m = x.shape[0]
    cdef int n = y.shape[0]
    if(m < 1 or n < 1):
        raise ValueError('Dynamic time warping can not handle empty input sequences!')
    cdef int K = x.shape[1]
    if(y.shape[1] != K):
        raise ValueError('x and y do not have the same dimensionality (%d versus %d)' % (x.shape[1], y.shape[1]))
    # First, compute all pairwise replacements
    # using OMP parallelization
    cdef int i
    cdef int j
    Delta = np.zeros((m, n))
    cdef double[:,:] Delta_view = Delta
    cdef int k
    cdef double diff
    for i in prange(m, nogil=True):
        for j in prange(n):
            for k in prange(K):
                diff = x[i,k] - y[j,k]
                if(diff < 0):
                    Delta_view[i, j] -= diff
                else:
                    Delta_view[i, j] += diff
    # Then, compute the dynamic time warping distance
    D = np.zeros((m,n))
    dtw_c(Delta, D)
    return D[0,0]

@cython.boundscheck(False)
def dtw_euclidean(double[:,:] x, double[:,:] y):
    """ Computes the multivariate dynamic time warping distance between two
    input arrays x and y, using the Euclidean distance as element-wise
    distance measure.

    Parameters
    ----------
    x: array_like
        a m x K matrix of doubles.
    y: array_like
        a n x K matrix of doubles.

    Returns
    -------
    d: float
        the dynamic time warping distance between x and y.

    """
    cdef int m = x.shape[0]
    cdef int n = y.shape[0]
    if(m < 1 or n < 1):
        raise ValueError('Dynamic time warping can not handle empty input sequences!')
    cdef int K = x.shape[1]
    if(y.shape[1] != K):
        raise ValueError('x and y do not have the same dimensionality (%d versus %d)' % (x.shape[1], y.shape[1]))
    # First, compute all pairwise replacements
    # using OMP parallelization
    cdef int i
    cdef int j
    Delta = np.zeros((m, n))
    cdef double[:,:] Delta_view = Delta
    cdef int k
    cdef double diff
    for i in prange(m, nogil=True):
        for j in prange(n):
            for k in prange(K):
                diff = x[i,k] - y[j,k]
                Delta_view[i, j] += diff * diff
            Delta_view[i, j] = sqrt(Delta_view[i, j])
    # Then, compute the dynamic time warping distance
    D = np.zeros((m,n))
    dtw_c(Delta, D)
    return D[0,0]

@cython.boundscheck(False)
def dtw_string(str x, str y):
    """ Computes the dynamic time warping distance between two
    input strings x and y, using the Kronecker distance as element-wise
    distance measure.

    Parameters
    ----------
    x: str
        a string.
    y: str
        another string.

    Returns
    -------
    d: float
        the dynamic time warping distance between x and y.

    """
    cdef int m = len(x)
    cdef int n = len(y)
    if(m < 1 or n < 1):
        raise ValueError('Dynamic time warping can not handle empty input sequences!')
    # First, compute all pairwise replacements
    cdef int i
    cdef int j
    Delta = np.zeros((m, n))
    cdef double[:,:] Delta_view = Delta
    for i in prange(m, nogil=True):
        for j in prange(n):
            if(x[i] == y[j]):
                Delta_view[i, j] = 0.
            else:
                Delta_view[i, j] = 1.
    # Then, compute the dynamic time warping distance
    D = np.zeros((m,n))
    dtw_c(Delta, D)
    return D[0,0]

@cython.boundscheck(False)
cdef void dtw_c(const double[:,:] Delta, double[:,:] D) nogil:
    """ Computes the dynamic time warping distance between two input sequences
    with pairwise element distances Delta and an (empty) dynamic programming
    matrix D.

    Parameters
    ----------
    Delta: array_like
        a m x n matrix containing the pairwise element distances.
    D: array:like
        another m x n matrix to which the output will be written.
        The dynamic time warping distance will be in cell [0, 0] after the
        computation is finished.

    """
    cdef int i
    cdef int j
    # initialize last entry
    D[-1, -1] = Delta[-1, -1]
    # compute last column
    for i in range(D.shape[0]-2,-1,-1):
        D[i,-1] = Delta[i,-1] + D[i+1,-1]
    # compute last row
    for j in range(D.shape[1]-2,-1,-1):
        D[-1,j] = Delta[-1,j] + D[-1,j+1]
    # compute remaining matrix
    for i in range(D.shape[0]-2,-1,-1):
        for j in range(D.shape[1]-2,-1,-1):
            D[i,j] = Delta[i,j] + min3(D[i+1,j+1], D[i,j+1], D[i+1,j])

cdef double min3(double a, double b, double c) nogil:
    """ Computes the minimum of three numbers.

    Parameters
    ----------
    a: double
        a number
    b: double
        another number
    c: double
        yet another number

    Returns
    -------
    min3: double
        min({a, b, c})

    """
    if(a < b):
        if(a < c):
            return a
        else:
            return c
    else:
        if(b < c):
            return b
        else:
            return c

####### BACKTRACING FUNCTIONS #######

cdef double _BACKTRACE_TOL = 1E-5

def dtw_backtrace(x, y, delta):
    """ Computes a co-optimal alignment between the two input sequences
    x and y, given the element-wise distance function delta. This mechanism
    is deterministic and will always prefer replacements over other options.

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.
    delta: function
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to dynamic time
        warping.

    """
    cdef int m = len(x)
    cdef int n = len(y)
    if(m < 1 or n < 1):
        raise ValueError('Dynamic time warping can not handle empty input sequences!')
    # First, compute all pairwise replacements
    Delta = np.zeros((m, n))
    cdef double[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            Delta_view[i,j] = delta(x[i], y[j])

    # Then, compute the dynamic time warping distance
    D = np.zeros((m,n))
    dtw_c(Delta, D)

    cdef double[:,:] D_view = D
    # Finally, compute the backtrace
    i = 0
    j = 0
    alignment = Alignment()
    while(i < m - 1 and j < n - 1):
        alignment.append_tuple(i, j)
        # check which alignment option is co-optimal
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j+1]):
            # replacement is co-optimal
            i += 1
            j += 1
            continue
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j]):
            # copying y[j] is co-optimal
            i += 1
            continue
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i,j+1]):
            # copying x[i] is co-optimal
            j += 1
            continue
        # if we got here, nothing is co-optimal, which is an error
        raise ValueError('Internal error: No option is co-optimal.')
    while(i < m - 1):
        alignment.append_tuple(i, j)
        i += 1
    while(j < n - 1):
        alignment.append_tuple(i, j)
        j += 1
    alignment.append_tuple(m-1, n-1)
    return alignment

def dtw_backtrace_stochastic(x, y, delta):
    """ Computes a co-optimal alignment between the two input sequences
    x and y, given the element-wise distance function delta. This mechanism
    is stochastic and will return a random co-optimal alignment.

    Note that the randomness does _not_ produce a uniform distribution over
    all co-optimal alignments because reandom choices at the start of the
    alignment process dominate. If you wish to characterize the overall
    distribution accurately, use sed_backtrace_matrix instead.

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.
    delta: function
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to dynamic time
        warping.

    """
    cdef int m = len(x)
    cdef int n = len(y)
    if(m < 1 or n < 1):
        raise ValueError('Dynamic time warping can not handle empty input sequences!')
    # First, compute all pairwise replacements
    Delta = np.zeros((m, n))
    cdef double[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            Delta_view[i,j] = delta(x[i], y[j])

    # Then, compute the dynamic time warping distance
    D = np.zeros((m,n))
    dtw_c(Delta, D)

    cdef double[:,:] D_view = D
    # Finally, compute the backtrace
    cdef int r
    i = 0
    j = 0
    alignment = Alignment()
    while(i < m - 1 and j < n - 1):
        alignment.append_tuple(i, j)
        # check which alignment options are co-optimal
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j+1]):
            # replacement is co-optimal
            if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j]):
                # replacement and copying y[j] are co-optimal
                if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i,j+1]):
                    # replacement, copying y[j], and copying x[i] are co-optimal
                    # Select whether to proceed in any direction uniformly at random
                    r = random.randrange(3)
                    if(r == 0):
                        i += 1
                        j += 1
                    elif(r == 1):
                        i += 1
                    else:
                        j += 1
                else:
                    # select whether to proceed in j direction according to a
                    # coin toss
                    i += 1
                    j += random.randrange(2)
            elif(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i,j+1]):
                # replacement and copying x[i] are co-optimal
                # select whether to proceed in i direction according to a
                # coin toss
                i += random.randrange(2)
                j += 1
            else:
                # only replacement is co-optimal
                i += 1
                j += 1
        elif(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j]):
            # copying y[j] is co-optimal
            if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i,j+1]):
                # copying y[j] and copying x[i] are co-optimal
                # Select whether to proceed in i or j direction uniformly at random
                r = random.randrange(2)
                if(r == 0):
                    i += 1
                else:
                    j += 1
            else:
                # only copying y[j] is co-optimal
                i += 1
        elif(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i,j+1]):
            # only copying x[i] is co-optimal
            j += 1
        else:
            # if we got here, nothing is co-optimal, which is an error
            raise ValueError('Internal error: No option is co-optimal.')
    while(i < m - 1):
        alignment.append_tuple(i, j)
        i += 1
    while(j < n - 1):
        alignment.append_tuple(i, j)
        j += 1
    alignment.append_tuple(m-1, n-1)
    return alignment

def dtw_backtrace_matrix(x, y, delta):
    """ Computes a matrix, summarizing all co-optimal alignments between
    x and y in a matrix P, where entry P[i, j] specifies the fraction of
    co-optimal alignments in which node x[i] has been aligned with node y[j].

    Parameters
    ----------
    x: list
        a sequence of objects.
    y: list
        another sequence of objects.
    delta: function
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them.

    Returns
    -------
    P: array_like
        a matrix, where entry P[i, j] specifies the fraction of co-optimal
        alignments in which node x[i] has been aligned with node y[j].
    K: array_like
        a matrix that contains the counts for all co-optimal alignments in
        which node x[i] has been aligned with node y[j].
    k: int
        the number of co-optimal alignments overall, such that P = K / k.

    """
    cdef int m = len(x)
    cdef int n = len(y)
    if(m < 1 or n < 1):
        raise ValueError('Dynamic time warping can not handle empty input sequences!')
    # First, compute all pairwise replacements
    Delta = np.zeros((m, n))
    cdef double[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            Delta_view[i,j] = delta(x[i], y[j])

    # Then, compute the dynamic time warping distance
    D = np.zeros((m,n))
    dtw_c(Delta, D)
    cdef double[:,:] D_view = D

    # compute the forward matrix Alpha, which contains the number of
    # co-optimal alignment paths from cell [0, 0] to cell [i, j]
    Alpha = np.zeros((m, n), dtype=int)
    cdef long[:,:] Alpha_view = Alpha
    Alpha_view[0, 0] = 1
    # build a queue of cells which we still need to process
    q = [(0, 0)]
    # build a set which stores the already visited cells
    visited = set()
    # initialize temporary variables
    cdef int found_coopt = False
    cdef long k = 0
    while(q):
        (i, j) = heapq.heappop(q)
        if((i, j) in visited):
            continue
        visited.add((i, j))
        k = Alpha_view[i, j]
        if(i == m-1):
            if(j == n-1):
                continue
            # if we are at the end of the first sequence, we can only copy
            # that end
            Alpha_view[i, j+1] += k
            heapq.heappush(q, (i, j+1))
            continue
        if(j == n-1):
            # if we are at the end of the second sequence, we can only copy
            # that end
            Alpha_view[i+1, j] += k
            heapq.heappush(q, (i+1, j))
            continue
        found_coopt = False
        # check which alignment option is co-optimal
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j+1]):
            # replacement is co-optimal
            Alpha_view[i+1, j+1] += k
            heapq.heappush(q, (i+1, j+1))
            found_coopt = True
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i+1,j]):
            # copying y[j] is co-optimal
            Alpha_view[i+1, j] += k
            heapq.heappush(q, (i+1, j))
            found_coopt = True
        if(D_view[i,j] + _BACKTRACE_TOL > Delta_view[i,j] + D_view[i,j+1]):
            # copying x[i] is co-optimal
            Alpha_view[i, j+1] += k
            heapq.heappush(q, (i, j+1))
            found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    # compute the backward matrix Beta, which contains the number of
    # co-optimal alignment paths from cell [i, j] to cell [m-1, n-1]
    Beta = np.zeros((m, n), dtype=int)
    cdef long[:,:] Beta_view = Beta
    Beta_view[m-1, n-1] = 1
    # iterate in downward lexigraphic order over the visited cells
    for (i, j) in sorted(visited, reverse = True):
        k = Beta_view[i, j]
        if(i == 0):
            if(j == 0):
                continue
            # if we are at the start of the first sequence, we can only copy
            # this start
            Beta_view[i, j-1] += k
            continue
        if(j == 0):
            # if we are at the start of the second sequence, we can only copy
            # this start
            Beta_view[i-1, j] += k
            continue
        found_coopt = False
        # check which alignment option is co-optimal
        if(D_view[i-1,j-1] + _BACKTRACE_TOL > Delta_view[i-1,j-1] + D_view[i,j]):
            # replacement is co-optimal
            Beta_view[i-1, j-1] += k
            found_coopt = True
        if(D_view[i-1,j] + _BACKTRACE_TOL > Delta_view[i-1,j] + D_view[i,j]):
            # copying y[j] is co-optimal
            Beta_view[i-1, j] += k
            found_coopt = True
        if(D_view[i,j-1] + _BACKTRACE_TOL > Delta_view[i,j-1] + D_view[i,j]):
            # copying x[i] is co-optimal
            Beta_view[i, j-1] += k
            found_coopt = True
        if(not found_coopt):
            raise ValueError('Internal error: No option is co-optimal.')

    if(Alpha_view[m-1, n-1] != Beta_view[0, 0]):
        raise ValueError('Internal error: Alignment count in Alpha and Beta matrix did not agree; got %d versus %d' % (Alpha_view[m-1, n-1], Beta_view[0, 0]))

    # compute a counting matrix specifying how often each alignment has
    # occured by multiplying alpha and beta values.
    K = Alpha * Beta
    # compute the final summary matrix by dividing K by the overall number
    # of co-optimal alignments
    return K.astype(np.float) / Alpha_view[m-1, n-1], K, Alpha_view[m-1, n-1]




#!python
#cython: language_level=3
"""
Provides a set edit distance to compare two sets (each represented as
lists for computational ease).

"""

# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

cimport cython
import numpy as np
from scipy.optimize import linear_sum_assignment
from edist.alignment import Alignment

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

def seted(x, y, delta = None):
    """ Computes the set edit distance between the input set x and the input
    set y, given the element-wise distance function delta.

    In more detail, this function finds an alignment M = {(i, j)} subset of
    {1, ..., len(x)} x {1, ..., len(y)}, such that the following loss is
    minimized:

    .. math:: \\sum_{(i, j) \\in M} \\delta(x_i, y_j) + \\sum_{i \\notin M} \\delta(x_i, -) + \\sum_{j \\notin M} \\delta(-, y_j)

    where i is said to be not in M if there exists no tuple (i, j) in M for
    any j, and j is said to be not in M if there exists no tuple (i, j) in M
    for any i.

    This problem can be solved via the Hungarian or Munkres algorithm in
    O([len(x) + len(y)]^3). So be advised that this method can become very slow
    for large input sets.

    Parameters
    ----------
    x: list-like
        a set of objects.
    y: list-like
        another set of objects.
    delta: function (default = None)
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them. If not given,
        standard_seted is called instead.

    Returns
    -------
    d: float
        the set edit distance between x and y according to delta.

    """
    if(delta is None):
        return standard_seted(x, y)

    d, J = _seted(x, y, delta)
    return d

def _seted(x, y, delta):
    """ Internal function, use seted instead.
    """

    cdef int m = len(x)
    cdef int n = len(y)
    # compute all replacement, deletion, and insertion costs
    Delta = np.zeros((m+1, n+1))
    cdef double[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            Delta_view[i, j] = delta(x[i], y[j])
    for i in range(m):
        Delta_view[i, n] = delta(x[i], None)
    for j in range(n):
        Delta_view[m, j] = delta(None, y[j])
    # compute the set edit distance
    return _seted_matrix(Delta)

def _seted_matrix(double[:, :] Delta):
    """ Computes the set edit distance between the input set x and the input
    set y, given the pairwise replacement costs as well as deletion and
    insertion costs.

    In more detail, this function finds an alignment M = {(i, j)} subset of
    {1, ..., m} x {1, ..., n}, where m = len(x) and n = len(y), such that the
    following loss is minimized:

    .. math:: \\sum_{(i, j) \\in M} \\Delta_{i, j} + \\sum_{i \\notin M} \\Delta_{i, n} + \\sum_{j \\notin M} \\Delta_{m, j}

    where i is said to be not in M if there exists no tuple (i, j) in M for
    any j, and j is said to be not in M if there exists no tuple (i, j) in M
    for any i.

    This problem can be solved via the Hungarian or Munkres algorithm in
    O([len(x) + len(y)]^3). So be advised that this method can become very slow
    for large input sets.

    Parameters
    ----------
    Delta: double matrix
        A m + 1 x n + 1 matrix, where Delta[i, j] for i < m and j < n is the
        cost of replacing x[i] with y[j], where Delta[i, n] is the deletion
        cost for x[i], and where Delta[m, j] is the insertion cost for y[j].

    Returns
    -------
    d: float
        the set edit distance between x and y according to Delta.
    J: int array
        J[i] is the index j to which i is mapped. If J[i] is at least n, i
        is deleted. Inversely, if some index j is not contained in J, it is
        inserted.

    """
    # retrieve
    cdef int m = Delta.shape[0] - 1
    cdef int n = Delta.shape[1] - 1
    # set up a cost matrix C for the Hungarian algorithm
    C = np.full((m + n, m + n), np.inf)
    cdef double[:,:] C_view = C
    # set the submatrix 1:m, 1:n to the replacement costs
    C_view[:m, :n] = Delta[:m, :n]
    # set the alignment cost of i with n+i to the deletion costs
    cdef int i
    for i in range(m):
        C_view[i, n+i] = Delta[i, n]
    # set the alignment cost of m+j with j to the insertion costs
    cdef int j
    for j in range(n):
        C_view[m+j, j] = Delta[m, j]
    # set the alignment costs for m+j with n+i to zero
    C_view[m:, n:] = 0.
    # then, apply the Hungarian algorithm
    I, J = linear_sum_assignment(C)
    # and recover the edit distance
    d = np.sum(C[I, J])
    return d, J

def seted_backtrace(x, y, delta = None):
    """ Computes a co-optimal alignment between the two input sequences
    x and y, given the element-wise distance function delta. This mechanism
    is deterministic and depends on the implementation of
    scipy.optimize.linear_sum_assignment for the choice of co-optimal
    alignment.

    Parameters
    ----------
    x: list-like
        a set of objects.
    y: list-like
        another set of objects.
    delta: function (default = None)
        a function that takes an element of x as first and an element of y
        as second input and returns the distance between them. If None, this
        method calls standard_seted_backtrace instead.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the set edit
        distance.

    """
    if delta is None:
        return standard_seted_backtrace(x, y)

    d, J = _seted(x, y, delta)
    return _to_alignment(J, len(x), len(y))

def _to_alignment(long[:] J, int m, int n):
    """ Transforms a given mapping to an alignment object.

    Parameters
    ----------
    J: int array
        J[i] is the index j to which i is mapped. If J[i] is at least n, i
        is deleted. Inversely, if some index j is not contained in J, it is
        inserted.
    m: int
        the size of the first input set.
    n: int
        the size of the second input set.

    Returns
    -------
    alignment: class alignment.Alignment
        the alignment object corresponding to J.

    """
    alignment = Alignment()
    cdef int i
    for i in range(m):
        if J[i] < n:
            alignment.append_tuple(i, J[i])
        else:
            alignment.append_tuple(i, -1)
    cdef int j
    for j in range(n):
        if J[m+j] < n:
            alignment.append_tuple(-1, j)
    return alignment

def standard_seted(x, y):
    """ Computes the standard set edit distance between the input set x and
    the input set y according to the Kronecker distance, i.e. the replacement
    cost is zero if two elements are equal and one otherwise.

    In more detail, this function finds an alignment M = {(i, j)} subset of
    {1, ..., len(x)} x {1, ..., len(y)}, such that the following loss is
    minimized:

    .. math:: \\sum_{(i, j) \\in M} \\delta(x_i, y_j) + |{i \\notin M}| + |{j \\notin M}|

    where i is said to be not in M if there exists no tuple (i, j) in M for
    any j, and j is said to be not in M if there exists no tuple (i, j) in M
    for any i.

    This problem can be solved via the Hungarian or Munkres algorithm in
    O([len(x) + len(y)]^3). So be advised that this method can become very slow
    for large input sets.

    Parameters
    ----------
    x: list-like
        a set of objects.
    y: list-like
        another set of objects.

    Returns
    -------
    d: float
        the set edit distance between x and y according to delta.

    """
    d, J = _standard_seted(x, y)
    return d

def _standard_seted(x, y):
    """ Internal function, use standard_seted instead.
    """

    cdef int m = len(x)
    cdef int n = len(y)
    # compute all replacement, deletion, and insertion costs
    Delta = np.zeros((m+1, n+1))
    cdef double[:,:] Delta_view = Delta
    cdef int i
    cdef int j
    for i in range(m):
        for j in range(n):
            if x[i] != y[j]:
                Delta_view[i, j] = 1.
    Delta_view[:, n] = 1.
    Delta_view[m, :] = 1.
    return _seted_matrix(Delta)

def standard_seted_backtrace(x, y):
    """ Computes a co-optimal alignment between the two input sequences
    x and y, given the element-wise distance function delta. This mechanism
    is deterministic and depends on the implementation of
    scipy.optimize.linear_sum_assignment for the choice of co-optimal
    alignment.

    Parameters
    ----------
    x: list-like
        a set of objects.
    y: list-like
        another set of objects.

    Returns
    -------
    alignment: class alignment.Alignment
        A co-optimal alignment between x and y according to the standard
        set edit distance.

    """
    d, J = _standard_seted(x, y)
    return _to_alignment(J, len(x), len(y))



"""
Provides general utility functions to process trees.

"""
# Copyright (C) 2019-2021
# Benjamin Paaßen
# AG Machine Learning
# Bielefeld University

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import numpy as np

__author__ = 'Benjamin Paaßen'
__copyright__ = 'Copyright (C) 2019-2021, Benjamin Paaßen'
__license__ = 'GPLv3'
__version__ = '1.2.0'
__maintainer__ = 'Benjamin Paaßen'
__email__  = 'bpaassen@techfak.uni-bielefeld.de'

def root(adj):
    """ Returns the root of a tree and raises an error if the input adjacency
    matrix does not correspond to a tree.

    Parameters
    ----------
    adj: list
        a directed graph in adjacency list format.

    Returns
    -------
    root: int
        the index of the root of the tree.

    Raises
    ------
    ValueError
        if the given adjacency list does not form a tree.

    """
    if(not adj):
        raise ValueError("The input tree is empty!")

    par = np.full(len(adj), -1, dtype=int)
    for i in range(len(adj)):
        for j in adj[i]:
            if(par[j] < 0):
                par[j] = i
            else:
                raise ValueError("Input is not a tree because node %d has multiple parents" % j)
    root = -1
    for i in range(len(adj)):
        if(par[i] < 0):
            if(root < 0):
                root = i
            else:
                raise ValueError("Input is not a tree because there is more than one root")
    return root

def check_tree_structure(adj):
    """ Verifies that a given adjacency list describes a tree.

    In particular, we build a parent representation of the tree and throw
    an exception if that fails. This is valid because trees are the subclass
    of directed graphs with a unique parent.

    Parameters
    ----------
    adj: list
        a directed graph in adjacency list format.

    Raises
    ------
    ValueError
        if the given adjacency list does not form a tree.

    """
    m = len(adj)
    # initialize an undefined parent array
    par = np.full(m, -1)
    # iterate through the adjacency list and retrieve the parents for each
    # node
    for i in range(m):
        # j iterates over all children of i
        for j in adj[i]:
            # if the parent of j is already defined, j has multiple parents
            # and this is not a tree
            if(par[j] >= 0):
                raise ValueError("Node %d has multiple parents (%d and %d)." % (j, par[j], i))
            par[j] = i
    return par


def check_dfs_structure(adj, i = 0):
    """ Verifies that a given adjacency list is in depth-first-search order,
    in the sense that the descendants of a node i are all indices i+1, i+2,
    ... orl[i], where orl[i] is the outermost right leaf of i.

    We verify this by performing a depth-first search over the tree and
    checking whether the indices are equivalent

    Parameters
    ----------
    adj: list
        a directed graph in adjacency list format.

    Returns
    -------
    size: int
        the size of the input tree, i.e. the last DFS index.

    Raises
    ------
    ValueError
        if the given adjacency list is not in DFS format.

    """
    if(not adj):
        return 0
    # j iterates over all children of adj i
    for j in adj[i]:
        # check the current child
        if(j != i + 1):
            raise ValueError("Expected %d as next child, but got %d" % (i+1, j))
        # check the dfs structure of the current child
        i = check_dfs_structure(adj, i+1)
    # return the current dfs index
    return i


def to_dfs_structure(nodes, adj):
    """ Re-orders a tree to conform to a depth-first search structure.

    Note that this method performs a copy and leaves the original tree
    untouched.

    Parameters
    ----------
    nodes: list
        A node list.
    adj: list
        An adjacency list.

    Raises
    ------
    ValueError
        if the input is not a tree.

    """
    # initialize tree copy
    nodes_dfs = []
    adj_dfs = []
    # get the root of the tree
    r = root(adj)
    # then, perform a depth-first search through the tree
    # and order the tree accordingly
    _to_dfs_structure(nodes, adj, nodes_dfs, adj_dfs, r)
    return nodes_dfs, adj_dfs


def _to_dfs_structure(nodes_orig, adj_orig, nodes_dfs, adj_dfs, i):
    # append the current node to the tree
    nodes_dfs.append(nodes_orig[i])
    # generate a new adjacency list
    adj_i = []
    adj_dfs.append(adj_i)
    # re-order the children of the current node
    for j in adj_orig[i]:
        # add the current child to the new adjacency list
        adj_i.append(len(nodes_dfs))
        # re-order the current child
        _to_dfs_structure(nodes_orig, adj_orig, nodes_dfs, adj_dfs, j)

def to_json(filename, nodes, adj):
    """ Writes a tree in node list/adjacency list format to a JSON file.

    Parameters
    ----------
    filename: str
        The filename for the resulting JSON file.
    nodes: list
        The node list of the tree.
    adj: list
        The adjacency list of the tree.

    Raises
    ------
    Exception
        if the file is not accessible or the JSON writeout fails.

    """
    with open(filename, 'w') as json_file:
        json.dump({'nodes' : nodes, 'adj' : adj}, json_file, indent='\t')


def from_json(filename):
    """Loads a tree in node list/adjacency list format from a JSON file.

    Parameters
    ----------
    filename: str
        A JSON filename containing tree data as written by the to_json method.

    Returns
    -------
    nodes: list
        The node list of the tree.
    adj: list
        The adjacency list of the tree.

    Raises
    ------
    Exception
        if the given file is not accessible, if the JSON data is malformed,
        if the parsed data is not a tree, or if the parsed tree is not in
        depth first search order.

    """
    with open(filename, 'r') as json_file:
        tree = json.load(json_file)
        nodes = tree['nodes']
        adj = tree['adj']
        check_tree_structure(adj)
        check_dfs_structure(adj)
        return nodes, adj

def dataset_from_json(path):
    """ Reads trees in node list/adjacency list format from all JSON files in
    the given directory.

    Parameters
    ----------
    path: str
        a path to a directory which contains JSON files.

    Returns
    -------
    X: list
        A list of tuples in node list/adjacency list format.
    filenames: list
        A list of filenames from which we read the trees.

    Raises
    ------
    Exception
        if the file access does not work, if some JSON data is malformed,
        if a parsed data is not a tree, or if a parsed tree is not in depth
        first search order.

    """
    if(not os.path.isdir(path)):
        raise OSError('%s is not a directory or was not found' % str(path))

    # list all files in the directory and sort them.
    files = os.listdir(path)
    files.sort()
    # initialize an empty output list.
    X = []
    filenames = []
    # iterate over all files and parse them
    for filename in files:
        # note that 'listdir' returns files relative to the input path, so
        # we have to put the input path in front.
        filepath = path + '/' + filename
        # check if the file is a json file.
        if(not os.path.isfile(filepath) or not filename.endswith('.json')):
            continue
        # load the tree
        X.append(from_json(filepath))
        filenames.append(filename)
    return X, filenames

def tree_to_string(nodes, adj, indent = False, with_indices = False):
    """ Prints a tree in node list/adjacency list format as string.

    Parameters
    ----------
    nodes: list
        The node list of the tree.
    adj: list
        The adjacency list of the tree.
    indent: bool (default = False)
        A boolean flag; if True, each node is printed on a new line.
    with_indices: bool (default = False)
        A boolean flag; if True, each node is printed with its index.

    Raises
    ------
    ValueError
        if the adjacency list does not correspond to a tree.

    """
    r = root(adj)
    if(indent):
        indent = 1
    else:
        indent = None
    # translate recursively
    return _tree_to_string(nodes, adj, r, indent, with_indices = with_indices)

def _tree_to_string(nodes, adj, i, indent = None, with_indices = False):
    # the initial string is just the node label
    if(with_indices):
        tree_string = '%d: %s' % (i, str(nodes[i]))
    else:
        tree_string = str(nodes[i])
    # consider the case where node i has chidlren
    if(adj[i]):
        # first, translate the children to their tree strings
        children_strings = []
        for j in adj[i]:
            if(indent is None):
                children_strings.append(_tree_to_string(nodes, adj, j, with_indices = with_indices))
            else:
                children_strings.append(_tree_to_string(nodes, adj, j, indent + 1, with_indices = with_indices))
        # and then join all these strings and append them in brackets
        if(indent is None):
            tree_string += '(' + ', '.join(children_strings) + ')'
        else:
            tree_string += '(\n' + ('\t' * indent) + (',\n' + '\t' * indent).join(children_strings) + '\n' + ('\t' * (indent - 1)) + ')'
    return tree_string

def subtree(nodes, adj, i):
    """ Returns the subtree rooted at node i in node list/adjacency list
    format.

    Parameters
    ----------
    nodes: list
        The node list of the tree.
    adj: list
        The adjacency list of the tree.
    i: int
        The index of the desired subtree.

    Returns
    -------
    nodes_sub: list
        The node list of the subtree rooted at i.
    adj_sub: list
        The adajency list of the subtree rooted at i.

    """
    # perform a depth-first-search from node i to construct the tree.
    nodes_dfs = []
    adj_dfs = []
    _to_dfs_structure(nodes, adj, nodes_dfs, adj_dfs, i)
    return nodes_dfs, adj_dfs

def parents(adj):
    """ Returns the parent representation of the tree with the given
    adjacency list.

    Parameters
    ----------
    adj: list
        The adjacency list of the tree.

    Returns
    -------
    par: int array
        a numpy integer array with len(adj) elements, where the ith
        element contains the index of the parent of the ith node.
        Nodes without children contain the entry -1.

    """
    par = np.full(len(adj), -1, dtype=int)
    for i in range(len(adj)):
        for j in adj[i]:
            par[j] = i
    return par
